#     Copyright (c) 2018 Jouko Strömmer
#     Copyright (c) 2017 Waveshare
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.
from papertty.drivers.drivers_base import WaveshareEPD


class WavesharePartial(WaveshareEPD):
    """Displays that support partial refresh (*monochrome*): 1.54", 2.13", 2.9". 
    The code is almost entirely identical with these, just small differences in the 2.13"."""

    BOOSTER_SOFT_START_CONTROL = 0x0C
    BORDER_WAVEFORM_CONTROL = 0x3C
    DATA_ENTRY_MODE_SETTING = 0x11
    DEEP_SLEEP_MODE = 0x10
    DISPLAY_UPDATE_CONTROL_1 = 0x21
    DISPLAY_UPDATE_CONTROL_2 = 0x22
    DRIVER_OUTPUT_CONTROL = 0x01
    GATE_SCAN_START_POSITION = 0x0F
    MASTER_ACTIVATION = 0x20
    SET_DUMMY_LINE_PERIOD = 0x3A
    SET_GATE_TIME = 0x3B
    SET_RAM_X_ADDRESS_COUNTER = 0x4E
    SET_RAM_X_ADDRESS_START_END_POSITION = 0x44
    SET_RAM_Y_ADDRESS_COUNTER = 0x4F
    SET_RAM_Y_ADDRESS_START_END_POSITION = 0x45
    SW_RESET = 0x12
    TEMPERATURE_SENSOR_CONTROL = 0x1A
    TERMINATE_FRAME_READ_WRITE = 0xFF
    WRITE_LUT_REGISTER = 0x32
    WRITE_RAM = 0x24
    WRITE_VCOM_REGISTER = 0x2C

    # these LUTs are used by 1.54" and 2.9" - 2.13" overrides them
    lut_full_update = [
        0x02, 0x02, 0x01, 0x11, 0x12, 0x12, 0x22, 0x22,
        0x66, 0x69, 0x69, 0x59, 0x58, 0x99, 0x99, 0x88,
        0x00, 0x00, 0x00, 0x00, 0xF8, 0xB4, 0x13, 0x51,
        0x35, 0x51, 0x51, 0x19, 0x01, 0x00
    ]

    lut_partial_update = [
        0x10, 0x18, 0x18, 0x08, 0x18, 0x18, 0x08, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x13, 0x14, 0x44, 0x12,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.supports_partial = True
        self.colors = 2
        self.lut = None

    def init(self, partial=True):
        self.partial_refresh = partial
        if self.epd_init() != 0:
            return -1
        # EPD hardware init start
        self.lut = self.lut_partial_update if partial else self.lut_full_update
        self.reset()
        self.send_command(self.DRIVER_OUTPUT_CONTROL)
        self.send_data((self.height - 1) & 0xFF)
        self.send_data(((self.height - 1) >> 8) & 0xFF)
        self.send_data(0x00)  # GD = 0 SM = 0 TB = 0
        self.send_command(self.BOOSTER_SOFT_START_CONTROL)
        self.send_data(0xD7)
        self.send_data(0xD6)
        self.send_data(0x9D)
        self.send_command(self.WRITE_VCOM_REGISTER)
        self.send_data(0xA8)  # VCOM 7C
        self.send_command(self.SET_DUMMY_LINE_PERIOD)
        self.send_data(0x1A)  # 4 dummy lines per gate
        self.send_command(self.SET_GATE_TIME)
        self.send_data(0x08)  # 2us per line
        self.send_command(self.DATA_ENTRY_MODE_SETTING)
        self.send_data(0x03)  # X increment Y increment
        self.set_lut(self.lut)
        # EPD hardware init end
        return 0

    def wait_until_idle(self):
        while self.digital_read(self.BUSY_PIN) == 1:  # 0: idle, 1: busy
            self.delay_ms(100)

    def set_lut(self, lut):
        self.lut = lut
        self.send_command(self.WRITE_LUT_REGISTER)
        # the length of look-up table is 30 bytes
        for i in range(0, len(lut)):
            self.send_data(self.lut[i])

    def get_frame_buffer(self, image):
        buf = [0x00] * int(self.width * self.height / 8)
        # Set buffer to value of Python Imaging Library image.
        # Image must be in mode 1.
        image_monocolor = image.convert('1')
        imwidth, imheight = image_monocolor.size
        if imwidth != self.width or imheight != self.height:
            raise ValueError('Image must be same dimensions as display \
                ({0}x{1}).'.format(self.width, self.height))

        pixels = image_monocolor.load()
        for y in range(self.height):
            for x in range(self.width):
                # Set the bits for the column of pixels at the current position.
                if pixels[x, y] != 0:
                    buf[int((x + y * self.width) / 8)] |= 0x80 >> (x % 8)
        return buf

    # this differs with 2.13" but is the same for 1.54" and 2.9"
    def set_frame_memory(self, image, x, y):
        if image is None or x < 0 or y < 0:
            return
        image_monocolor = image.convert('1')
        image_width, image_height = image_monocolor.size
        # x point must be the multiple of 8 or the last 3 bits will be ignored
        x = x & 0xF8
        image_width = image_width & 0xF8
        if x + image_width >= self.width:
            x_end = self.width - 1
        else:
            x_end = x + image_width - 1
        if y + image_height >= self.height:
            y_end = self.height - 1
        else:
            y_end = y + image_height - 1
        self.set_memory_area(x, y, x_end, y_end)
        self.set_memory_pointer(x, y)
        self.send_command(self.WRITE_RAM)
        # send the image data
        pixels = image_monocolor.load()
        byte_to_send = 0x00
        for j in range(0, y_end - y + 1):
            # 1 byte = 8 pixels, steps of i = 8
            for i in range(0, x_end - x + 1):
                # Set the bits for the column of pixels at the current position.
                if pixels[i, j] != 0:
                    byte_to_send |= 0x80 >> (i % 8)
                if i % 8 == 7:
                    self.send_data(byte_to_send)
                    byte_to_send = 0x00

    def clear_frame_memory(self, color):
        self.set_memory_area(0, 0, self.width - 1, self.height - 1)
        self.set_memory_pointer(0, 0)
        self.send_command(self.WRITE_RAM)
        # send the color data
        for i in range(0, int(self.width / 8 * self.height)):
            self.send_data(color)

    def display_frame(self):
        self.send_command(self.DISPLAY_UPDATE_CONTROL_2)
        self.send_data(0xC4)
        self.send_command(self.MASTER_ACTIVATION)
        self.send_command(self.TERMINATE_FRAME_READ_WRITE)
        self.wait_until_idle()

    def set_memory_area(self, x_start, y_start, x_end, y_end):
        self.send_command(self.SET_RAM_X_ADDRESS_START_END_POSITION)
        # x point must be the multiple of 8 or the last 3 bits will be ignored
        self.send_data((x_start >> 3) & 0xFF)
        self.send_data((x_end >> 3) & 0xFF)
        self.send_command(self.SET_RAM_Y_ADDRESS_START_END_POSITION)
        self.send_data(y_start & 0xFF)
        self.send_data((y_start >> 8) & 0xFF)
        self.send_data(y_end & 0xFF)
        self.send_data((y_end >> 8) & 0xFF)

    def set_memory_pointer(self, x, y):
        self.send_command(self.SET_RAM_X_ADDRESS_COUNTER)
        # x point must be the multiple of 8 or the last 3 bits will be ignored
        self.send_data((x >> 3) & 0xFF)
        self.send_command(self.SET_RAM_Y_ADDRESS_COUNTER)
        self.send_data(y & 0xFF)
        self.send_data((y >> 8) & 0xFF)
        self.wait_until_idle()

    def sleep(self):
        self.send_command(self.DEEP_SLEEP_MODE)
        self.wait_until_idle()

    def draw(self, x, y, image):
        """Replace a particular area on the display with an image"""
        self.set_frame_memory(image, x, y)
        self.display_frame()
        if self.partial_refresh:
            # set the memory again if partial refresh LUT is used
            self.set_frame_memory(image, x, y)
            self.display_frame()


class EPD1in54(WavesharePartial):
    """Waveshare 1.54" - monochrome"""

    def __init__(self):
        super().__init__(name='1.54" BW', width=200, height=200)


class EPD2in13(WavesharePartial):
    """Waveshare 2.13" - monochrome"""

    lut_full_update = [
        0x22, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x11,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E,
        0x01, 0x00, 0x00, 0x00, 0x00, 0x00
    ]

    lut_partial_update = [
        0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x0F, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    ]

    def __init__(self):
        # the actual pixel width is 122, but 128 is the 'logical' width
        super().__init__(name='2.13" BW', width=128, height=250)

    def set_frame_memory(self, image, x, y):
        if image is None or x < 0 or y < 0:
            return
        image_monocolor = image.convert('1')
        image_width, image_height = image_monocolor.size
        # x point must be the multiple of 8 or the last 3 bits will be ignored
        x = x & 0xF8
        image_width = image_width & 0xF8
        if x + image_width >= self.width:
            x_end = self.width - 1
        else:
            x_end = x + image_width - 1
        if y + image_height >= self.height:
            y_end = self.height - 1
        else:
            y_end = y + image_height - 1
        self.set_memory_area(x, y, x_end, y_end)
        # send the image data
        pixels = image_monocolor.load()
        byte_to_send = 0x00
        for j in range(y, y_end + 1):
            self.set_memory_pointer(x, j)
            self.send_command(self.WRITE_RAM)
            # 1 byte = 8 pixels, steps of i = 8
            for i in range(x, x_end + 1):
                # Set the bits for the column of pixels at the current position.
                if pixels[i - x, j - y] != 0:
                    byte_to_send |= 0x80 >> (i % 8)
                if i % 8 == 7:
                    self.send_data(byte_to_send)
                    byte_to_send = 0x00


class EPD2in13v2(WavesharePartial):
    """Waveshare 2.13" V2 - monochrome"""

    lut_full_update = [
        0x80,0x60,0x40,0x00,0x00,0x00,0x00,             #LUT0: BB:     VS 0 ~7
        0x10,0x60,0x20,0x00,0x00,0x00,0x00,             #LUT1: BW:     VS 0 ~7
        0x80,0x60,0x40,0x00,0x00,0x00,0x00,             #LUT2: WB:     VS 0 ~7
        0x10,0x60,0x20,0x00,0x00,0x00,0x00,             #LUT3: WW:     VS 0 ~7
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,             #LUT4: VCOM:   VS 0 ~7

        0x03,0x03,0x00,0x00,0x02,                       # TP0 A~D RP0
        0x09,0x09,0x00,0x00,0x02,                       # TP1 A~D RP1
        0x03,0x03,0x00,0x00,0x02,                       # TP2 A~D RP2
        0x00,0x00,0x00,0x00,0x00,                       # TP3 A~D RP3
        0x00,0x00,0x00,0x00,0x00,                       # TP4 A~D RP4
        0x00,0x00,0x00,0x00,0x00,                       # TP5 A~D RP5
        0x00,0x00,0x00,0x00,0x00,                       # TP6 A~D RP6

        0x15,0x41,0xA8,0x32,0x30,0x0A
    ]

    lut_partial_update = [
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,             #LUT0: BB:     VS 0 ~7
        0x80,0x00,0x00,0x00,0x00,0x00,0x00,             #LUT1: BW:     VS 0 ~7
        0x40,0x00,0x00,0x00,0x00,0x00,0x00,             #LUT2: WB:     VS 0 ~7
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,             #LUT3: WW:     VS 0 ~7
        0x00,0x00,0x00,0x00,0x00,0x00,0x00,             #LUT4: VCOM:   VS 0 ~7

        0x0A,0x00,0x00,0x00,0x00,                       # TP0 A~D RP0
        0x00,0x00,0x00,0x00,0x00,                       # TP1 A~D RP1
        0x00,0x00,0x00,0x00,0x00,                       # TP2 A~D RP2
        0x00,0x00,0x00,0x00,0x00,                       # TP3 A~D RP3
        0x00,0x00,0x00,0x00,0x00,                       # TP4 A~D RP4
        0x00,0x00,0x00,0x00,0x00,                       # TP5 A~D RP5
        0x00,0x00,0x00,0x00,0x00,                       # TP6 A~D RP6

        0x15,0x41,0xA8,0x32,0x30,0x0A,
    ]

    def __init__(self):
        # the actual pixel width is 122, but 128 is the 'logical' width
        super().__init__(name='2.13" BW V2 (full refresh only)', width=128, height=250)


class EPD2in9(WavesharePartial):
    """Waveshare 2.9" - monochrome"""

    def __init__(self):
        super().__init__(name='2.9" BW', width=128, height=296)


class EPD2in13d(WavesharePartial):
    """Waveshare 2.13" D - monochrome (flexible)"""

    # Note: the original code for this display was pretty broken and seemed
    # to have been written by some other person than the rest of the drivers.

    def __init__(self):
        super().__init__(name='2.13" D', width=104, height=212)

    lut_vcomDC = [
        0x00, 0x08, 0x00, 0x00, 0x00, 0x02,
        0x60, 0x28, 0x28, 0x00, 0x00, 0x01,
        0x00, 0x14, 0x00, 0x00, 0x00, 0x01,
        0x00, 0x12, 0x12, 0x00, 0x00, 0x01,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00,
    ]

    lut_ww = [
        0x40, 0x08, 0x00, 0x00, 0x00, 0x02,
        0x90, 0x28, 0x28, 0x00, 0x00, 0x01,
        0x40, 0x14, 0x00, 0x00, 0x00, 0x01,
        0xA0, 0x12, 0x12, 0x00, 0x00, 0x01,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    ]

    lut_bw = [
        0x40, 0x17, 0x00, 0x00, 0x00, 0x02,
        0x90, 0x0F, 0x0F, 0x00, 0x00, 0x03,
        0x40, 0x0A, 0x01, 0x00, 0x00, 0x01,
        0xA0, 0x0E, 0x0E, 0x00, 0x00, 0x02,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    ]

    lut_wb = [
        0x80, 0x08, 0x00, 0x00, 0x00, 0x02,
        0x90, 0x28, 0x28, 0x00, 0x00, 0x01,
        0x80, 0x14, 0x00, 0x00, 0x00, 0x01,
        0x50, 0x12, 0x12, 0x00, 0x00, 0x01,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    ]

    lut_bb = [
        0x80, 0x08, 0x00, 0x00, 0x00, 0x02,
        0x90, 0x28, 0x28, 0x00, 0x00, 0x01,
        0x80, 0x14, 0x00, 0x00, 0x00, 0x01,
        0x50, 0x12, 0x12, 0x00, 0x00, 0x01,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    ]

    lut_vcom1 = [
        0x00, 0x19, 0x01, 0x00, 0x00, 0x01,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00,
    ]

    lut_ww1 = [
        0x00, 0x19, 0x01, 0x00, 0x00, 0x01,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    ]

    lut_bw1 = [
        0x80, 0x19, 0x01, 0x00, 0x00, 0x01,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    ]

    lut_wb1 = [
        0x40, 0x19, 0x01, 0x00, 0x00, 0x01,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    ]

    lut_bb1 = [
        0x00, 0x19, 0x01, 0x00, 0x00, 0x01,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    ]

    def wait_until_idle(self):
        """This particular model's code sends the GET_STATUS command while waiting - dunno why."""
        while self.digital_read(self.BUSY_PIN) == 0:  # 0: idle, 1: busy
            self.send_command(self.GET_STATUS)
            self.delay_ms(100)

    def init(self, **kwargs):
        if self.epd_init() != 0:
            return -1
        self.reset()

        self.send_command(0x01)  # POWER SETTING
        self.send_data(0x03)
        self.send_data(0x00)
        self.send_data(0x2b)
        self.send_data(0x2b)
        self.send_data(0x03)

        self.send_command(0x06)  # boost soft start
        self.send_data(0x17)  # A
        self.send_data(0x17)  # B
        self.send_data(0x17)  # C

        self.send_command(0x04)
        self.wait_until_idle()

        self.send_command(0x00)  # panel setting
        self.send_data(0xbf)  # LUT from OTP,128x296
        self.send_data(0x0d)  # VCOM to 0V fast

        self.send_command(0x30)  # PLL setting
        self.send_data(0x3a)  # 3a 100HZ   29 150Hz 39 200HZ	31 171HZ

        self.send_command(0x61)  # resolution setting
        self.send_data(self.width)
        self.send_data((self.height >> 8) & 0xff)
        self.send_data(self.height & 0xff)

        self.send_command(0x82)  # vcom_DC setting
        self.send_data(0x28)

        # self.send_command(0X50)			#VCOM AND DATA INTERVAL SETTING
        # self.send_data(0xb7)		#WBmode:VBDF 17|D7 VBDW 97 VBDB 57		WBRmode:VBDF F7 VBDW 77 VBDB 37  VBDR B7
        return 0

    def set_full_reg(self):
        self.send_command(0x82)
        self.send_data(0x00)
        self.send_command(0X50)
        self.send_data(0xb7)

        self.send_command(0x20)  # vcom
        for count in range(0, 44):
            self.send_data(self.lut_vcomDC[count])
        self.send_command(0x21)  # ww --
        for count in range(0, 42):
            self.send_data(self.lut_ww[count])
        self.send_command(0x22)  # bw r
        for count in range(0, 42):
            self.send_data(self.lut_bw[count])
        self.send_command(0x23)  # wb w
        for count in range(0, 42):
            self.send_data(self.lut_wb[count])
        self.send_command(0x24)  # bb b
        for count in range(0, 42):
            self.send_data(self.lut_bb[count])

    def set_part_reg(self):
        self.send_command(0x82)
        self.send_data(0x03)
        self.send_command(0X50)
        self.send_data(0x47)

        self.send_command(0x20)  # vcom
        for count in range(0, 44):
            self.send_data(self.lut_vcom1[count])
        self.send_command(0x21)  # ww --
        for count in range(0, 42):
            self.send_data(self.lut_ww1[count])
        self.send_command(0x22)  # bw r
        for count in range(0, 42):
            self.send_data(self.lut_bw1[count])
        self.send_command(0x23)  # wb w
        for count in range(0, 42):
            self.send_data(self.lut_wb1[count])
        self.send_command(0x24)  # bb b
        for count in range(0, 42):
            self.send_data(self.lut_bb1[count])

    def turn_on_display(self):
        self.send_command(0x12)
        self.delay_ms(10)
        self.wait_until_idle()

    def clear(self):
        self.send_command(0x10)
        for i in range(0, int(self.width * self.height / 8)):
            self.send_data(0x00)
        self.delay_ms(10)

        self.send_command(0x13)
        for i in range(0, int(self.width * self.height / 8)):
            self.send_data(0xFF)
        self.delay_ms(10)

        self.set_full_reg()
        self.turn_on_display()

    def display_full(self, frame_buffer):
        if not frame_buffer:
            return

        self.send_command(0x10)
        for i in range(0, int(self.width * self.height / 8)):
            self.send_data(0x00)
        self.delay_ms(10)

        self.send_command(0x13)
        for i in range(0, int(self.width * self.height / 8)):
            self.send_data(frame_buffer[i])
        self.delay_ms(10)

        self.set_full_reg()
        self.turn_on_display()

    def display_partial(self, frame_buffer, x_start, y_start, x_end, y_end):
        if not frame_buffer:
            return

        self.set_part_reg()
        self.send_command(0x91)
        self.send_command(0x90)
        self.send_data(x_start)
        self.send_data(x_end - 1)

        self.send_data(y_start / 256)
        self.send_data(y_start % 256)
        self.send_data(y_end / 256)
        self.send_data(y_end % 256 - 1)
        self.send_data(0x28)

        self.send_command(0x10)
        for i in range(0, int(self.width * self.height / 8)):
            # print(frame_buffer[i],'%d','0x10')
            self.send_data(frame_buffer[i])
        self.delay_ms(10)

        self.send_command(0x13)
        for i in range(0, int(self.width * self.height / 8)):
            # print(~frame_buffer[i],'%d','0x13')
            self.send_data(~frame_buffer[i])
        self.delay_ms(10)

        # self.set_full_reg()
        self.turn_on_display()

    # after this, call epd.init() to awaken the module
    def sleep(self):
        self.send_command(0x50)
        self.send_data(0xf7)
        self.send_command(0x02)  # power off
        self.send_command(0x07)  # deep sleep
        self.send_data(0xA5)

    def draw(self, x, y, image):
        """Replace a particular area on the display with an image"""
        if self.partial_refresh:
            self.display_partial(self.get_frame_buffer(image), x, y, x + image.width, x + image.height)
        else:
            self.display_full(self.get_frame_buffer(image))



#     Copyright (c) 2020 Guido Kraemer
#     Copyright (c) 2018 Jouko Strömmer
#     Copyright (c) 2017 Waveshare
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from PIL import Image

from papertty.drivers.drivers_consts import EPD4in2const
from papertty.drivers.drivers_partial import WavesharePartial

try:
    import RPi.GPIO as GPIO
except ImportError:
    pass
except RuntimeError as e:
    print(str(e))

# The driver works as follows:
#
# This driver is for the following display:
#     400x300, 4.2inch E-Ink display module
#     SKU: 13353
#     Part Number: 4.2inch e-Paper Module
#     Brand: Waveshare
#     UPC: 614961950887
#
# When rotating the display with side of the connector of the display (not the
# module) at the bottom, the display width is 400 and the height is 300. The
# letters on the module and the connector of the module point to the right.
#
# The origin of the display is in the top left corner and filling happens by
# line.
#
# The frame buffer is an array of bytes of size 400 * 300 / 8, each bit is one
# pixel 0 is white, 1 is black
#
# The framebuffer should always contain the entire image, redrawing happens
# only on the canged area.
#
# The properties width and height of the incoming image correspond to the
# properties of the display, access to the pixels of the image is:
# image.load[width, height]


class EPD4in2(WavesharePartial, EPD4in2const):
    """WaveShare 4.2" """

    # code adapted from  epd_4in2.c
    # https://github.com/waveshare/e-paper/blob/8973995e53cb78bac6d1f8a66c2d398c18392f71/raspberrypi%26jetsonnano/c/lib/e-paper/epd_4in2.c

    # note: this works differently (at least in the c code): there is a memory
    # buffer, the same size as the display. we partially refresh the memory
    # buffer with the image at a position and the do self.partial_refresh with
    # the entire memory buffer and the area to be refreshed.

    # note: this code is outside of drivers_partial.py because the class has to
    # override many methdos and therefore is way to long

    def __init__(self):
        super(WavesharePartial, self).__init__(name='4.2"',
                                               width=400,
                                               height=300)
        self.supports_partial = True

        # this is the memory buffer that will be updated!
        self.frame_buffer = [0x00] * (self.width * self.height // 8)

    # TODO: universal?
    def set_setting(self, command, data):
        self.send_command(command)
        self.send_data(data)
        # for d in data:
        #     self.send_data(d)

    # TODO: universal?
    def set_resolution(self):
        self.set_setting(self.RESOLUTION_SETTING,
                         [(self.width >> 8) & 0xff,
                          self.width & 0xff,
                          (self.height >> 8) & 0xff,
                          self.height & 0xff])

    def reset(self):
        self.digital_write(self.RST_PIN, GPIO.HIGH)
        self.delay_ms(200)
        self.digital_write(self.RST_PIN, GPIO.LOW)
        self.delay_ms(200)
        self.digital_write(self.RST_PIN, GPIO.HIGH)
        self.delay_ms(200)

    def send_command(self, command):
        self.digital_write(self.DC_PIN, GPIO.LOW)
        self.digital_write(self.CS_PIN, GPIO.LOW)
        self.spi_transfer([command])
        self.digital_write(self.CS_PIN, GPIO.HIGH)

    def send_data(self, data):
        self.digital_write(self.DC_PIN, GPIO.HIGH)
        self.digital_write(self.CS_PIN, GPIO.LOW)
        if type(data) == list:
            self.spi_transfer(data)
        else:
            self.spi_transfer([data])
        self.digital_write(self.CS_PIN, GPIO.HIGH)

    def wait_until_idle(self):
        self.send_command(self.GET_STATUS)
        while self.digital_read(self.BUSY_PIN) == 0:
            self.send_command(self.GET_STATUS)
            self.delay_ms(100)

    def turn_on_display(self):
        self.send_command(self.DISPLAY_REFRESH)
        self.delay_ms(100)
        self.wait_until_idle()

    def full_set_lut(self):
        self.set_setting(self.VCOM_LUT, self.LUT_VCOM0)
        self.set_setting(self.W2W_LUT, self.LUT_WW)
        self.set_setting(self.B2W_LUT, self.LUT_BW)
        self.set_setting(self.W2B_LUT, self.LUT_WB)
        self.set_setting(self.B2B_LUT, self.LUT_BB)

    def partial_set_lut(self):
        self.set_setting(self.VCOM_LUT, self.PARTIAL_LUT_VCOM1)
        self.set_setting(self.W2W_LUT, self.PARTIAL_LUT_WW1)
        self.set_setting(self.B2W_LUT, self.PARTIAL_LUT_BW1)
        self.set_setting(self.W2B_LUT, self.PARTIAL_LUT_WB1)
        self.set_setting(self.B2B_LUT, self.PARTIAL_LUT_BB1)

    def gray_set_lut(self):
        self.set_setting(self.VCOM_LUT, self.GRAY_LUT_VCOM)
        self.set_setting(self.W2W_LUT, self.GRAY_LUT_WW)
        self.set_setting(self.B2W_LUT, self.GRAY_LUT_BW)
        self.set_setting(self.W2B_LUT, self.GRAY_LUT_WB)
        self.set_setting(self.B2B_LUT, self.GRAY_LUT_BB)

    def init_bw(self):
        self.reset()
        self.set_setting(self.POWER_SETTING, [0x03, 0x00, 0x2b, 0x2b])
        self.set_setting(self.BOOSTER_SOFT_START, [0x17, 0x17, 0x17])
        self.send_command(self.POWER_ON)
        self.wait_until_idle()
        # kw-bf kwr-af bwrotp 0f bwotp 1f
        self.set_setting(self.PANEL_SETTING, [0xbf, 0x0d])
        # 3a 100hz   29 150hz 39 200hz	31 171hz
        self.set_setting(self.PLL_CONTROL, [0x3c])
        self.set_resolution()
        self.set_setting(self.VCM_DC_SETTING, [0x28])
        # wbmode: vbdf 17|d7 vbdw 97 vbdb 57
        # wbrmode:vbdf f7 vbdw 77 vbdb 37 vbdr b7
        self.set_setting(self.VCOM_AND_DATA_INTERVAL_SETTING, [0x97])
        self.full_set_lut()

    def init_gray(self):
        self.reset()
        self.set_setting(self.POWER_SETTING, [0x03, 0x00, 0x2b, 0x2b, 0x13])
        self.set_setting(self.BOOSTER_SOFT_START, [0x17, 0x17, 0x17])
        self.send_command(self.POWER_ON)
        self.wait_until_idle()
        # kw-3f  kwr-2f  bwrotp 0f  bwotp 1f
        self.set_setting(self.PANEL_SETTING, [0x3f])
        # 3a 100hz   29 150hz 39 200hz	31 171hz
        self.set_setting(self.PLL_CONTROL, [0x3c])
        self.set_resolution()
        self.set_setting(self.VCM_DC_SETTING, [0x12])
        # wbmode:vbdf 17|d7 vbdw 97 vbdb 57
        # wbrmode:vbdf f7 vbdw 77 vbdb 37 vbdr b7
        self.set_setting(self.VCOM_AND_DATA_INTERVAL_SETTING, [0x97])

    def init(self, partial=True, gray=False):
        self.partial_refresh = partial
        self.gray = gray

        if self.epd_init() != 0:
            return -1

        if self.gray:
            self.init_gray()
        else:
            self.init_bw()

    def clear(self):
        """Full refresh of the display"""

        width = int(self.width // 8)
        height = int(self.height)

        self.send_command(self.DATA_START_TRANSMISSION_1)
        for i in range(width * height):
            self.send_data(0xff)

        self.send_command(self.DATA_START_TRANSMISSION_2)
        for i in range(width * height):
            self.send_data(0xff)

        self.send_command(self.DISPLAY_REFRESH)
        self.delay_ms(10)
        self.turn_on_display()

    # Writing outside the range of the display will cause an error.
    def fill(self, color, fillsize):
        """Slow fill routine"""

        div, rem = divmod(self.height, fillsize)
        image = Image.new('1', (self.width, fillsize), color)

        for i in range(div):
            self.draw(0, i * fillsize, image)

        if rem != 0:
            image = Image.new('1', (self.width, rem), color)
            self.draw(0, div * fillsize, image)

    def display_full(self):

        # TODO: optimize this as (needs performance check):
        # self.send_command(self.DATA_START_TRANSMISSION_2, self.frame_buffer)

        width = int(self.width // 8)
        height = int(self.height)

        self.send_command(self.DATA_START_TRANSMISSION_2)
        for i in range(height):
            for j in range(width):
                self.send_data(self.frame_buffer[i * width + j])

        self.turn_on_display()

    def display_partial(self, x_start, y_start, x_end, y_end):

        width = self.width // 8

        # TODO: use x_start // 8 * 9
        x_start = int(x_start if x_start % 8 == 0 else x_start // 8 * 8 + 8)
        x_end = int(x_end if x_end % 8 == 0 else x_end // 8 * 8 + 8)

        self.set_setting(self.VCOM_AND_DATA_INTERVAL_SETTING, [0xf7])
        self.delay_ms(100)

        self.set_setting(self.VCM_DC_SETTING, [0x08])
        self.set_setting(self.VCOM_AND_DATA_INTERVAL_SETTING, [0x47])
        self.partial_set_lut()

        self.send_command(self.PARTIAL_IN)
        self.set_setting(self.PARTIAL_WINDOW,
                         [x_start     // 256, x_start     % 256,
                          (x_end - 1) // 256, (x_end - 1) % 256,
                          y_start     // 256, y_start     % 256,
                          (y_end - 1) // 256, (y_end - 1) % 256,
                          0x28])

        # writes old data to sram for programming
        self.send_command(self.DATA_START_TRANSMISSION_1)
        for j in range(y_end - y_start):
            idx = (y_start + j) * width + x_start // 8
            # TODO: optimize this by using send_data(array[from:to])
            for i in range((x_end - x_start) // 8):
                self.send_data(self.frame_buffer[idx + i])

        # writes new data to sram.
        self.send_command(self.DATA_START_TRANSMISSION_2)
        for j in range(y_end - y_start):
            idx = (y_start + j) * width + x_start // 8
            # TODO: optimize this by using send_data(array[from:to])
            for i in range((x_end - x_start) // 8):
                self.send_data(~self.frame_buffer[idx + i])

        self.send_command(self.DISPLAY_REFRESH)   # display refresh
        self.delay_ms(10)  # the delay here is necessary, 200us at least!!!
        self.turn_on_display()

    def sleep(self):
        """Put the display in deep sleep mode"""

        self.send_command(self.POWER_OFF)
        self.wait_until_idle()
        self.set_setting(self.DEEP_SLEEP, [0xa5])

    def frame_buffer_to_image(self):
        """Returns self.frame_buffer as a PIL.Image"""

        im = Image.new('1', (self.width, self.height), "white")
        pi = im.load()

        for j in range(self.height):
            idxj = j * self.width // 8
            for i in range(self.width):
                idiv, irem = divmod(i, 8)
                mask = 0b10000000 >> irem
                idxi = idiv
                pi[i, j] = self.frame_buffer[idxi + idxj] & mask

        return im

    def set_frame_buffer(self, x, y, image):
        """Updates self.frame_buffer with image at (x, y)"""

        image_monocolor = image.convert('1')
        imwidth, imheight = image_monocolor.size
        pixels = image_monocolor.load()

        for j in range(imheight):
            idxj = (y + j) * self.width // 8
            for i in range(imwidth):
                idiv, irem = divmod(x + i, 8)
                mask = 0b10000000 >> irem
                idxi = idiv
                if pixels[i, j] != 0:
                    self.frame_buffer[idxi + idxj] |= mask
                else:
                    self.frame_buffer[idxi + idxj] &= ~mask

    def draw(self, x, y, image):
        """replace a particular area on the display with an image"""

        if self.partial_refresh:
            self.set_frame_buffer(x, y, image)
            self.display_partial(x, y, x + image.width, y + image.height)
        else:
            self.set_frame_buffer(0, 0, image)
            self.display_full()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Colin Nolan, 2020
# Jouko Strömmer, 2018
# Copyright and related rights waived via CC0
# https://creativecommons.org/publicdomain/zero/1.0/legalcode

# As you would expect, use this at your own risk! This code was created
# so you (yes, YOU!) can make it better.
#
# Requires Python 3

# display drivers - note: they are GPL licensed, unlike this file
import papertty.drivers.drivers_base as drivers_base
import papertty.drivers.drivers_partial as drivers_partial
import papertty.drivers.drivers_full as drivers_full
import papertty.drivers.drivers_color as drivers_color
import papertty.drivers.drivers_colordraw as drivers_colordraw
import papertty.drivers.driver_it8951 as driver_it8951
import papertty.drivers.drivers_4in2 as driver_4in2

# for ioctl
import fcntl
# for validating type of and access to device files
import os
# for gracefully handling signals (systemd service)
import signal
# for unpacking virtual console data
import struct
# for stdin and exit
import sys
import select
# for setting TTY size
import termios
# for sleeping
import time
# for command line usage
import click
# for drawing
from PIL import Image, ImageChops, ImageDraw, ImageFont, ImageOps
# for tidy driver list
from collections import OrderedDict
# for VNC
from vncdotool import api
# for reading stdin data for use with Pillow
from io import BytesIO

# resource path
RESOURCE_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "resources")

class PaperTTY:
    """The main class - handles various settings and showing text on the display"""
    defaultfont = os.path.join(RESOURCE_PATH, "tom-thumb.pil")
    defaultsize = 8
    driver = None
    partial = None
    initialized = None
    font = None
    fontsize = None
    font_height = None
    font_width = None
    white = None
    black = None
    encoding = None
    spacing = 0
    cursor = None
    rows = None
    cols = None
    is_truetype = None
    fontfile = None

    def __init__(self, driver, font=defaultfont, fontsize=defaultsize, partial=None, encoding='utf-8', spacing=0, cursor=None):
        """Create a PaperTTY with the chosen driver and settings"""
        self.driver = get_drivers()[driver]['class']()
        self.spacing = spacing
        self.fontsize = fontsize
        self.font = self.load_font(font) if font else None
        self.partial = partial
        self.white = self.driver.white
        self.black = self.driver.black
        self.encoding = encoding
        self.cursor = cursor

    def ready(self):
        """Check that the driver is loaded and initialized"""
        return self.driver and self.initialized

    @staticmethod
    def error(msg, code=1):
        """Print error and exit"""
        print(msg)
        sys.exit(code)

    def set_tty_size(self, tty, rows, cols):
        """Set a TTY (/dev/tty*) to a certain size. Must be a real TTY that support ioctls."""
        self.rows = rows
        self.cols = cols
        with open(tty, 'w') as tty:
            size = struct.pack("HHHH", int(rows), int(cols), 0, 0)
            try:
                fcntl.ioctl(tty.fileno(), termios.TIOCSWINSZ, size)
            except OSError:
                print("TTY refused to resize (rows={}, cols={}), continuing anyway.".format(rows, cols))
                print("Try setting a sane size manually.")

    @staticmethod
    def band(bb):
        """Stretch a bounding box's X coordinates to be divisible by 8,
           otherwise weird artifacts occur as some bits are skipped."""
        return (int(bb[0] / 8) * 8, bb[1], int((bb[2] + 7) / 8) * 8, bb[3]) if bb else None

    @staticmethod
    def split(s, n):
        """Split a sequence into parts of size n"""
        return [s[begin:begin + n] for begin in range(0, len(s), n)]

    @staticmethod
    def fold(text, width=None, filter_fn=None):
        """Format a string to a specified width and/or filter it"""
        buff = text
        if width:
            buff = ''.join([r + '\n' for r in PaperTTY.split(buff, int(width))]).rstrip()
        if filter_fn:
            buff = [c for c in buff if filter_fn(c)]
        return buff

    @staticmethod
    def img_diff(img1, img2):
        """Return the bounding box of differences between two images"""
        return ImageChops.difference(img1, img2).getbbox()

    @staticmethod
    def ttydev(vcsa):
        """Return associated tty for vcsa device, ie. /dev/vcsa1 -> /dev/tty1"""
        return vcsa.replace("vcsa", "tty")
    
    def vcsudev(self, vcsa):
        """Return character width and associated vcs(u) for vcsa device,
           ie. for /dev/vcsa1, retunr (4, "/dev/vcsu1") if vcsu is available, or
           (1, "/dev/vcs1") if not"""
        dev = vcsa.replace("vcsa", "vcsu")
        if os.path.exists(dev):
            if isinstance(self.font, ImageFont.FreeTypeFont):
                return 4, dev
            else:
                print("Font {} doesn't support Unicode. Falling back to 8-bit encoding.".format(self.font.file))
                return 1, vcsa.replace("vcsa", "vcs")
        else:
            print("System does not have /dev/vcsu. Falling back to 8-bit encoding.")
            return 1, vcsa.replace("vcsa", "vcs")

    @staticmethod
    def valid_vcsa(vcsa):
        """Check that the vcsa device and associated terminal seem sane"""
        vcsa_kernel_major = 7
        tty_kernel_major = 4
        vcsa_range = range(128, 191)
        tty_range = range(1, 63)

        tty = PaperTTY.ttydev(vcsa)
        vs = os.stat(vcsa)
        ts = os.stat(tty)

        vcsa_major, vcsa_minor = os.major(vs.st_rdev), os.minor(vs.st_rdev)
        tty_major, tty_minor = os.major(ts.st_rdev), os.minor(ts.st_rdev)
        if not (vcsa_major == vcsa_kernel_major and vcsa_minor in vcsa_range):
            print("Not a valid vcsa device node: {} ({}/{})".format(vcsa, vcsa_major, vcsa_minor))
            return False
        read_vcsa = os.access(vcsa, os.R_OK)
        write_tty = os.access(tty, os.W_OK)
        if not read_vcsa:
            print("No read access to {} - maybe run with sudo?".format(vcsa))
            return False
        if not (tty_major == tty_kernel_major and tty_minor in tty_range):
            print("Not a valid TTY device node: {}".format(vcsa))
        if not write_tty:
            print("No write access to {} so cannot set terminal size, maybe run with sudo?".format(tty))
        return True

    def load_font(self, path, keep_if_not_found=False):
        """Load the PIL or TrueType font"""
        font = None
        # If no path is given, reuse existing font path. Good for resizing.
        path = path or self.fontfile
        if os.path.isfile(path):
            try:
                # first check if the font looks like a PILfont
                with open(path, 'rb') as f:
                    if f.readline() == b"PILfont\n":
                        self.is_truetype = False
                        print('Loading PIL font {}. Font size is ignored.'.format(path))
                        font = ImageFont.load(path)
                        # otherwise assume it's a TrueType font
                    else:
                        self.is_truetype = True
                        font = ImageFont.truetype(path, self.fontsize)
                    self.fontfile = path
            except IOError:
                self.error("Invalid font: '{}'".format(path))
        elif keep_if_not_found:
            print("The font '{}' could not be found, keep using old font.".format(path))
            font = self.font
        else:
            print("The font '{}' could not be found, using fallback font instead.".format(path))
            font = ImageFont.load_default()

        if font:
            # get physical dimensions of font. Take the average width of
            # 1000 M's because oblique fonts are complicated.
            self.font_width = font.getsize('M' * 1000)[0] // 1000
            if 'getmetrics' in dir(font):
                metrics_ascent, metrics_descent = font.getmetrics()
                self.spacing = int(self.spacing) if self.spacing != 'auto' else (metrics_descent - 2)
                print('Setting spacing to {}.'.format(self.spacing))
                # despite what the PIL docs say, ascent appears to be the
                # height of the font, while descent is not, in fact, negative.
                # Couuld use testing with more fonts.
                self.font_height = metrics_ascent + self.spacing
            else:
                # No autospacing for pil fonts, but they usually don't need it.
                self.spacing = int(self.spacing) if self.spacing != 'auto' else 0
                # pil fonts don't seem to have metrics, but all
                # characters seem to have the same height
                self.font_height = font.getsize('a')[1] + self.spacing

        return font

    def recalculate_font(self):
        """Load the PIL or TrueType font"""
        # get physical dimensions of font. Take the average width of
        # 1000 M's because oblique fonts a complicated.
        self.font_width = self.font.getsize('M' * 1000)[0] // 1000
        if 'getmetrics' in dir(self.font):
            metrics_ascent, metrics_descent = self.font.getmetrics()
            self.spacing = int(self.spacing) if self.spacing != 'auto' else (metrics_descent - 2)
            print('Setting spacing to {}.'.format(self.spacing))
            # despite what the PIL docs say, ascent appears to be the
            # height of the font, while descent is not, in fact, negative.
            # Couuld use testing with more fonts.
            self.font_height = metrics_ascent + self.spacing
        else:
            # No autospacing for pil fonts, but they usually don't need it.
            self.spacing = int(self.spacing) if self.spacing != 'auto' else 0
            # pil fonts don't seem to have metrics, but all
            # characters seem to have the same height
            self.font_height = self.font.getsize('a')[1] + self.spacing

    def init_display(self):
        """Initialize the display - call the driver's init method"""
        self.driver.init(partial=self.partial)
        self.initialized = True

    def fit(self, portrait=False):
        """Return the maximum columns and rows we can display with this font"""
        width = self.font.getsize('M')[0]
        height = self.font_height
        # hacky, subtract just a bit to avoid going over the border with small fonts
        pw = self.driver.width - 3
        ph = self.driver.height
        return int((pw if portrait else ph) / width), int((ph if portrait else pw) / height)

    def draw_line_cursor(self, cursor, draw):
        cur_x, cur_y = cursor[0], cursor[1]
        width = self.font_width
        # desired cursor width
        cur_width = width - 1
        # get font height
        height = self.font_height
        # starting X is the font width times current column
        start_x = cur_x * width
        offset = 0
        if self.cursor != 'default': # only default and a number are valid in this context
            offset = int(self.cursor)
        # add 1 because rows start at 0 and we want the cursor at the bottom
        start_y = (cur_y + 1) * height - 1 - offset
        # draw the cursor line
        draw.line((start_x, start_y, start_x + cur_width, start_y), fill=self.black)

    def draw_block_cursor(self, cursor, image):
        cur_x, cur_y = cursor[0], cursor[1]
        width = self.font_width
        # get font height
        height = self.font_height
        upper_left = (cur_x * width, cur_y * height)
        lower_right = ((cur_x + 1) * width, (cur_y + 1) * height)
        mask = Image.new('1', (image.width, image.height), self.black)
        draw = ImageDraw.Draw(mask)
        draw.rectangle([upper_left, lower_right], fill=self.white)
        return ImageChops.logical_xor(image, mask)

    def showfb(self, fb_num, rotate=None, invert=False, sleep=1, full_interval=100):
        """Render the framebuffer - basically a copy-paste of showvnc at this point"""
        def _get_fb_info(fb_num):
            config_dir = "/sys/class/graphics/fb%d/" % fb_num
            size = None
            bpp = None
            with open(config_dir + "/virtual_size", "r") as f:
                size = tuple([int(t) for t in f.read().strip().split(",")])
            with open(config_dir + "/bits_per_pixel", "r") as f:
                bpp = int(f.read().strip())
            return (size,bpp)

        def _get_fb_img(fb_num):
            size, bpp = _get_fb_info(fb_num)
            with open("/dev/fb%d" % fb_num, "rb") as f:
                mode = "BGRX" if bpp == 32 else "BGR;16"
                return Image.frombytes("RGB", size, f.read(), "raw", mode).convert("L")

        previous_fb_img = None
        diff_bbox = None
        # number of updates; when it's 0, do a full refresh
        updates = 0
        while True:
            new_fb_img = _get_fb_img(fb_num)
            # apply rotation if any
            if rotate:
                new_fb_img = new_fb_img.rotate(rotate, expand=True)
            # apply invert
            if invert:
                new_fb_img = ImageOps.invert(new_fb_img)
            # rescale image if needed
            if new_fb_img.size != (self.driver.width, self.driver.height):
                new_fb_img = new_fb_img.resize((self.driver.width, self.driver.height))
            # if at least two frames have been processed, get a bounding box of their difference region
            if new_fb_img and previous_fb_img:
                diff_bbox = self.band(self.img_diff(new_fb_img, previous_fb_img))
            # frames differ, so we should update the display
            if diff_bbox:
                # increment update counter
                updates = (updates + 1) % full_interval
                # if partial update is supported and it's not time for a full refresh,
                # draw just the different region
                if updates > 0 and (self.driver.supports_partial and self.partial):
                    print("partial ({}): {}".format(updates, diff_bbox))
                    self.driver.draw(diff_bbox[0], diff_bbox[1], new_fb_img.crop(diff_bbox))
                # if partial update is not possible or desired, do a full refresh
                else:
                    print("full ({}): {}".format(updates, new_fb_img.size))
                    old_partial = self.partial
                    self.partial = False
                    self.driver.draw(0, 0, new_fb_img)
                    self.partial = old_partial
            # otherwise this is the first frame, so run a full refresh to get things going
            else:
                if updates == 0:
                    updates = (updates + 1) % full_interval
                    print("initial ({}): {}".format(updates, new_fb_img.size))
                    self.driver.draw(0, 0, new_fb_img)
            previous_fb_img = new_fb_img.copy()
            time.sleep(float(sleep))



    def showvnc(self, host, display, password=None, rotate=None, invert=False, sleep=1, full_interval=100):
        with api.connect(':'.join([host, display]), password=password) as client:
            previous_vnc_image = None
            diff_bbox = None
            # number of updates; when it's 0, do a full refresh
            updates = 0
            client.timeout = 30
            while True:
                try:
                    client.refreshScreen()
                except TimeoutError:
                    print("Timeout to server {}:{}".format(host, display))
                    client.disconnect()
                    sys.exit(1)
                new_vnc_image = client.screen
                # apply rotation if any
                if rotate:
                    new_vnc_image = new_vnc_image.rotate(rotate, expand=True)
                # apply invert
                if invert:
                    new_vnc_image = ImageOps.invert(new_vnc_image)
                # rescale image if needed
                if new_vnc_image.size != (self.driver.width, self.driver.height):
                    new_vnc_image = new_vnc_image.resize((self.driver.width, self.driver.height))
                # if at least two frames have been processed, get a bounding box of their difference region
                if new_vnc_image and previous_vnc_image:
                    diff_bbox = self.band(self.img_diff(new_vnc_image, previous_vnc_image))
                # frames differ, so we should update the display
                if diff_bbox:
                    # increment update counter
                    updates = (updates + 1) % full_interval
                    # if partial update is supported and it's not time for a full refresh,
                    # draw just the different region
                    if updates > 0 and (self.driver.supports_partial and self.partial):
                        print("partial ({}): {}".format(updates, diff_bbox))
                        self.driver.draw(diff_bbox[0], diff_bbox[1], new_vnc_image.crop(diff_bbox))
                    # if partial update is not possible or desired, do a full refresh
                    else:
                        print("full ({}): {}".format(updates, new_vnc_image.size))
                        self.driver.draw(0, 0, new_vnc_image)
                # otherwise this is the first frame, so run a full refresh to get things going
                else:
                    if updates == 0:
                        updates = (updates + 1) % full_interval
                        print("initial ({}): {}".format(updates, new_vnc_image.size))
                        self.driver.draw(0, 0, new_vnc_image)
                previous_vnc_image = new_vnc_image.copy()
                time.sleep(float(sleep))

    def showtext(self, text, fill, cursor=None, portrait=False, flipx=False, flipy=False, oldimage=None):
        """Draw a string on the screen"""
        if self.ready():
            # set order of h, w according to orientation
            image = Image.new('1', (self.driver.width, self.driver.height) if portrait else (
                self.driver.height, self.driver.width),
                              self.white)
            # create the Draw object and draw the text
            draw = ImageDraw.Draw(image)
            draw.text((0, 0), text, font=self.font, fill=fill, spacing=self.spacing)

            # if we want a cursor, draw it - the most convoluted part
            if cursor and self.cursor:
                if self.cursor == 'block':
                    image = self.draw_block_cursor(cursor, image)
                else:
                    self.draw_line_cursor(cursor, draw)
            # rotate image if using landscape
            if not portrait:
                image = image.rotate(90, expand=True)
            # apply flips if desired
            if flipx:
                image = image.transpose(Image.FLIP_LEFT_RIGHT)
            if flipy:
                image = image.transpose(Image.FLIP_TOP_BOTTOM)
            # find out which part changed and draw only that on the display
            if oldimage and self.driver.supports_partial and self.partial:
                # create a bounding box of the altered region and
                # make the X coordinates divisible by 8
                diff_bbox = self.band(self.img_diff(image, oldimage))
                # crop the altered region and draw it on the display
                if diff_bbox:
                    self.driver.draw(diff_bbox[0], diff_bbox[1], image.crop(diff_bbox))
            else:
                # if no previous image, draw the entire display
                self.driver.draw(0, 0, image)
            return image
        else:
            self.error("Display not ready")
    
    def clear(self):
        """Clears the display; set all black, then all white, or use INIT mode, if driver supports it."""
        if self.ready():
            self.driver.clear()
            print('Display reinitialized.')
        else:
            self.error("Display not ready")


class Settings:
    """A class to store CLI settings so they can be referenced in the subcommands"""
    args = {}

    def __init__(self, **kwargs):
        self.args = kwargs

    def get_init_tty(self):
        tty = PaperTTY(**self.args)
        tty.init_display()
        return tty


def get_drivers():
    """Get the list of available drivers as a dict
    Format: { '<NAME>': { 'desc': '<DESCRIPTION>', 'class': <CLASS> }, ... }"""
    driverdict = {}
    driverlist = [drivers_partial.EPD1in54, drivers_partial.EPD2in13,
                  drivers_partial.EPD2in13v2, drivers_partial.EPD2in9,
                  drivers_partial.EPD2in13d, driver_4in2.EPD4in2,

                  drivers_full.EPD2in7, drivers_full.EPD3in7, drivers_full.EPD7in5,
                  drivers_full.EPD7in5v2,

                  drivers_color.EPD4in2b, drivers_color.EPD7in5b,
                  drivers_color.EPD5in83, drivers_color.EPD5in83b,
                  drivers_color.EPD5in65f,

                  drivers_colordraw.EPD1in54b, drivers_colordraw.EPD1in54c,
                  drivers_colordraw.EPD2in13b, drivers_colordraw.EPD2in7b,
                  drivers_colordraw.EPD2in9b,

                  driver_it8951.IT8951,

                  drivers_base.Dummy, drivers_base.Bitmap]
    for driver in driverlist:
        driverdict[driver.__name__] = {'desc': driver.__doc__, 'class': driver}
    return driverdict


def get_driver_list():
    """Get a neat printable driver list"""
    order = OrderedDict(sorted(get_drivers().items()))
    return '\n'.join(["{}{}".format(driver.ljust(15), order[driver]['desc']) for driver in order])


def display_image(driver, image, stretch=False, no_resize=False, fill_color="white", rotate=None, mirror=None, flip=None):
    """
    Display the given image using the given driver and options.
    :param driver: device driver (subclass of `WaveshareEPD`)
    :param image: image data to display
    :param stretch: whether to stretch the image so that it fills the screen in both dimentions
    :param no_resize: whether the image should not be resized if it does not fit the screen (will raise `RuntimeError`
    if image is too large)
    :param fill_color: colour to fill space when image is resized but one dimension does not fill the screen
    :param rotate: rotate the image by arbitrary degrees
    :param mirror: flip the image horizontally
    :param flip: flip the image vertically
    :return: the image that was rendered
    """
    if stretch and no_resize:
        raise ValueError('Cannot set "no-resize" with "stretch"')

    if mirror:
        image = ImageOps.mirror(image)
    if flip:
        image = ImageOps.flip(image)
    if rotate:
        image = image.rotate(rotate, expand=True, fillcolor=fill_color)
        
    image_width, image_height = image.size

    if stretch:
        if (image_width, image_height) == (driver.width, driver.height):
            output_image = image
        else:
            output_image = image.resize((driver.width, driver.height))
    else:
        if no_resize:
            if image_width > driver.width or image_height > driver.height:
                raise RuntimeError('Image ({0}x{1}) needs to be resized to fit the screen ({2}x{3})'
                                   .format(image_width, image_height, driver.width, driver.height))
            # Pad only
            output_image = Image.new(image.mode, (driver.width, driver.height), color=fill_color)
            output_image.paste(image, (0, 0))
        else:
            # Scales and pads
            output_image = ImageOps.pad(image, (driver.width, driver.height), color=fill_color)

    driver.draw(0, 0, output_image)

    return output_image


@click.group()
@click.option('--driver', default=None, help='Select display driver')
@click.option('--nopartial', is_flag=True, default=False, help="Don't use partial updates even if display supports it")
@click.option('--encoding', default='latin_1', help='Encoding to use for the buffer', show_default=True)
@click.pass_context
def cli(ctx, driver, nopartial, encoding):
    """Display stdin or TTY on a Waveshare e-Paper display"""
    if not driver:
        PaperTTY.error(
            "You must choose a display driver. If your 'C' variant is not listed, use the 'B' driver.\n\n{}".format(
                get_driver_list()))
    else:
        matched_drivers = [n for n in get_drivers() if n.lower() == driver.lower()]
        if not matched_drivers:
            PaperTTY.error('Invalid driver selection, choose from:\n{}'.format(get_driver_list()))
        ctx.obj = Settings(driver=matched_drivers[0], partial=not nopartial, encoding=encoding)
    pass


@click.command(name='list')
def list_drivers():
    """List available display drivers"""
    PaperTTY.error(get_driver_list(), code=0)


@click.command()
@click.option('--size', default=16, help='Stripe size to fill with (8-32)')
@click.pass_obj
def scrub(settings, size):
    """Slowly fill with black, then white"""
    if size not in range(8, 32 + 1):
        PaperTTY.error("Invalid stripe size, must be 8-32")
    ptty = settings.get_init_tty()
    ptty.driver.scrub(fillsize=size)


@click.command()
@click.option('--font', default=PaperTTY.defaultfont, help='Path to a TrueType or PIL font',
              show_default=True)
@click.option('--size', 'fontsize', default=8, help='Font size', show_default=True)
@click.option('--width', default=None, help='Fit to width [default: display width / font width]')
@click.option('--portrait', default=False, is_flag=True, help='Use portrait orientation', show_default=True)
@click.option('--nofold', default=False, is_flag=True, help="Don't fold the input", show_default=True)
@click.option('--spacing', default='0', help='Line spacing for the text, "auto" to automatically determine a good value', show_default=True)
@click.pass_obj
def stdin(settings, font, fontsize, width, portrait, nofold, spacing):
    """Display standard input and leave it on screen"""
    settings.args['font'] = font
    settings.args['fontsize'] = fontsize
    settings.args['spacing'] = spacing
    ptty = settings.get_init_tty()
    text = sys.stdin.read()
    if not nofold:
        if width:
            text = ptty.fold(text, width)
        else:
            font_width = ptty.font.getsize('M')[0]
            max_width = int((ptty.driver.width - 8) / font_width) if portrait else int(ptty.driver.height / font_width)
            text = ptty.fold(text, width=max_width)
    ptty.showtext(text, fill=ptty.driver.black, portrait=portrait)


@click.command()
@click.option('--image', 'image_location', help='Location of image to display (omit for stdin)', show_default=True)
@click.option('--stretch', default=False, is_flag=True, show_default=True,
              help='Stretch image so that it fills the entire screen (may distort your image!)')
@click.option('--no-resize', default=False, is_flag=True, show_default=True,
              help='Do not resize image to fit the screen (an error will occur if the image is too large!)')
@click.option('--fill-color', default='white', help='Colour to pad image with', show_default=True)
@click.option('--mirror', default=False, is_flag=True, help='Mirror horizontally', show_default=True)
@click.option('--flip', default=False, is_flag=True, help='Mirror vertically', show_default=True)
@click.option('--rotate', default=0, help='Rotate the image by N degrees', show_default=True, type=float)
@click.pass_obj
def image(settings, image_location, stretch, no_resize, fill_color, mirror, flip, rotate):
    """ Display an image """
    if image_location is None or image_location == '-':
        # XXX: logging to stdout, in line with the rest of this project
        print('Reading image data from stdin... (set "--image" to load an image from a given file path)')
        image_data = BytesIO(sys.stdin.buffer.read())
        image = Image.open(image_data)
    else:
        image = Image.open(image_location)

    ptty = settings.get_init_tty()
    display_image(ptty.driver, image, stretch=stretch, no_resize=no_resize, fill_color=fill_color, rotate=rotate, mirror=mirror, flip=flip)


@click.command()
@click.option('--host', default="localhost", help="VNC host to connect to", show_default=True)
@click.option('--display', default="0", help="VNC display to use (0 = port 5900)", show_default=True)
@click.option('--password', default=None, help="VNC password")
@click.option('--rotate', default=None, help="Rotate screen (90 / 180 / 270)")
@click.option('--invert', default=False, is_flag=True, help="Invert colors")
@click.option('--sleep', default=1, show_default=True, help="Refresh interval (s)", type=float)
@click.option('--fullevery', default=50, show_default=True, help="# of partial updates between full updates")
@click.pass_obj
def vnc(settings, host, display, password, rotate, invert, sleep, fullevery):
    """Display a VNC desktop"""
    ptty = settings.get_init_tty()
    ptty.showvnc(host, display, password, int(rotate) if rotate else None, invert, sleep, fullevery)


@click.command()
@click.option('--fb-num', default="0", help="Framebuffer to display (/dev/fbX)", show_default=True)
@click.option('--rotate', default=None, help="Rotate screen (90 / 180 / 270)")
@click.option('--invert', default=False, is_flag=True, help="Invert colors")
@click.option('--sleep', default=1, show_default=True, help="Refresh interval (s)", type=float)
@click.option('--fullevery', default=50, show_default=True, help="# of partial updates between full updates")
@click.pass_obj
def fb(settings, fb_num, rotate, invert, sleep, fullevery):
    """Display the framebuffer"""
    ptty = settings.get_init_tty()
    ptty.showfb(int(fb_num), int(rotate) if rotate else None, invert, sleep, fullevery)


@click.command()
@click.option('--vcsa', default='/dev/vcsa1', help='Virtual console device (/dev/vcsa[1-63])', show_default=True)
@click.option('--font', default=PaperTTY.defaultfont, help='Path to a TrueType or PIL font', show_default=True)
@click.option('--size', 'fontsize', default=8, help='Font size', show_default=True)
@click.option('--noclear', default=False, is_flag=True, help='Leave display content on exit')
@click.option('--nocursor', default=False, is_flag=True, help="(DEPRECATED, use --cursor=none instead) Don't draw the cursor")
@click.option('--cursor', default='legacy', help='Set cursor type. Valid values are default (underscore cursor at a sensible place), block (inverts colors at cursor), none (draws no cursor) or a number n (underscore cursor n pixels from the bottom)', show_default=False)
@click.option('--sleep', default=0.1, help='Minimum sleep between refreshes', show_default=True)
@click.option('--rows', 'ttyrows', default=None, help='Set TTY rows (--cols required too)')
@click.option('--cols', 'ttycols', default=None, help='Set TTY columns (--rows required too)')
@click.option('--portrait', default=False, is_flag=True, help='Use portrait orientation', show_default=False)
@click.option('--flipx', default=False, is_flag=True, help='Flip X axis (EXPERIMENTAL/BROKEN)', show_default=False)
@click.option('--flipy', default=False, is_flag=True, help='Flip Y axis (EXPERIMENTAL/BROKEN)', show_default=False)
@click.option('--spacing', default='0', help='Line spacing for the text, "auto" to automatically determine a good value', show_default=True)
@click.option('--scrub', 'apply_scrub', is_flag=True, default=False, help='Apply scrub when starting up',
              show_default=True)
@click.option('--autofit', is_flag=True, default=False, help='Autofit terminal size to font size', show_default=True)
@click.option('--attributes', is_flag=True, default=False, help='Use attributes', show_default=True)
@click.option('--interactive', is_flag=True, default=False, help='Interactive mode')
@click.pass_obj
def terminal(settings, vcsa, font, fontsize, noclear, nocursor, cursor, sleep, ttyrows, ttycols, portrait, flipx, flipy,
             spacing, apply_scrub, autofit, attributes, interactive):
    """Display virtual console on an e-Paper display, exit with Ctrl-C."""
    settings.args['font'] = font
    settings.args['fontsize'] = fontsize
    settings.args['spacing'] = spacing

    if cursor != 'legacy' and nocursor:
        print("--cursor and --nocursor can't be used together. To hide the cursor, use --cursor=none")
        sys.exit(1)

    if nocursor:
        print("--nocursor is deprecated. Use --cursor=none instead")
        settings.args['cursor'] = None

    if cursor == 'default' or cursor == 'legacy':
        settings.args['cursor'] = 'default'
    elif cursor == 'none':
        settings.args['cursor'] = None
    else:
        settings.args['cursor'] = cursor

    ptty = settings.get_init_tty()

    if apply_scrub:
        ptty.driver.scrub()
    oldbuff = ''
    oldimage = None
    oldcursor = None
    # dirty - should refactor to make this cleaner
    flags = {'scrub_requested': False, 'show_menu': False, 'clear': False}
    
    # handle SIGINT from `systemctl stop` and Ctrl-C
    def sigint_handler(sig, frame):
        if not interactive:
            print("Exiting (SIGINT)...")
            if not noclear:
                ptty.showtext(oldbuff, fill=ptty.white, **textargs)
            sys.exit(0)
        else:
             print('Showing menu, please wait ...')
             flags['show_menu'] = True

    # toggle scrub flag when SIGUSR1 received
    def sigusr1_handler(sig, frame):
        print("Scrubbing display (SIGUSR1)...")
        flags['scrub_requested'] = True

    signal.signal(signal.SIGINT, sigint_handler)
    signal.signal(signal.SIGUSR1, sigusr1_handler)

    # group the various params for readability
    textargs = {'portrait': portrait, 'flipx': flipx, 'flipy': flipy}

    if any([ttyrows, ttycols]) and not all([ttyrows, ttycols]):
        ptty.error("You must define both --rows and --cols to change terminal size.")
    if ptty.valid_vcsa(vcsa):
        if all([ttyrows, ttycols]):
            ptty.set_tty_size(ptty.ttydev(vcsa), ttyrows, ttycols)
        else:
            # if size not specified manually, see if autofit was requested
            if autofit:
                max_dim = ptty.fit(portrait)
                print("Automatic resize of TTY to {} rows, {} columns".format(max_dim[1], max_dim[0]))
                ptty.set_tty_size(ptty.ttydev(vcsa), max_dim[1], max_dim[0])
        if interactive:
            print("Started displaying {}, minimum update interval {} s, open menu with Ctrl-C".format(vcsa, sleep))
        else:
            print("Started displaying {}, minimum update interval {} s, exit with Ctrl-C".format(vcsa, sleep))
        character_width, vcsudev = ptty.vcsudev(vcsa)
        while True:
            if flags['show_menu']:
                flags['show_menu'] = False
                print()
                print('Rendering paused. Enter')
                print('    (f) to change font,')
                print('    (s) to change spacing,')
                if ptty.is_truetype:
                    print('    (h) to change font size,')
                print('    (c) to scrub,')
                print('    (i) reinitialize display,')
                print('    (r) do a full refresh,')
                print('    (x) to exit,')
                print('    anything else to continue.')
                print('Command line arguments for current settings:\n    --font {} --size {} --spacing {}'.format(ptty.fontfile, ptty.fontsize, ptty.spacing))

                ch = sys.stdin.readline().strip()
                if ch == 'x':
                    if not noclear:
                        ptty.showtext(oldbuff, fill=ptty.white, **textargs)
                    sys.exit(0)
                elif ch == 'f':
                    print('Current font: {}'.format(ptty.fontfile))
                    new_font = click.prompt('Enter new font (leave empty to abort)', default='', show_default=False)
                    if new_font:
                        ptty.spacing = spacing
                        ptty.font = ptty.load_font(new_font, keep_if_not_found=True)
                        if autofit:
                            max_dim = ptty.fit(portrait)
                            print("Automatic resize of TTY to {} rows, {} columns".format(max_dim[1], max_dim[0]))
                            ptty.set_tty_size(ptty.ttydev(vcsa), max_dim[1], max_dim[0])
                        oldbuff = None
                    else:
                        print('Font not changed')
                elif ch == 's':
                    print('Current spacing: {}'.format(ptty.spacing))
                    new_spacing = click.prompt('Enter new spacing (leave empty to abort)', default='empty', type=int, show_default=False)
                    if new_spacing != 'empty':
                        ptty.spacing = new_spacing
                        ptty.recalculate_font()
                        if autofit:
                            max_dim = ptty.fit(portrait)
                            print("Automatic resize of TTY to {} rows, {} columns".format(max_dim[1], max_dim[0]))
                            ptty.set_tty_size(ptty.ttydev(vcsa), max_dim[1], max_dim[0])
                        oldbuff = None
                    else:
                        print('Spacing not changed')
                elif ch == 'h' and ptty.is_truetype:
                    print('Current font size: {}'.format(ptty.fontsize))
                    new_fontsize = click.prompt('Enter new font size (leave empty to abort)', default='empty', type=int, show_default=False)
                    if new_fontsize != 'empty':
                        ptty.fontsize = new_fontsize
                        ptty.spacing = spacing
                        ptty.font = ptty.load_font(path=None)
                        if autofit:
                            max_dim = ptty.fit(portrait)
                            print("Automatic resize of TTY to {} rows, {} columns".format(max_dim[1], max_dim[0]))
                            ptty.set_tty_size(ptty.ttydev(vcsa), max_dim[1], max_dim[0])
                        oldbuff = None
                    else:
                        print('Font size not changed')
                elif ch == 'c':
                    flags['scrub_requested'] = True
                elif ch == 'i':
                    ptty.clear()
                    oldimage = None
                    oldbuff = None
                elif ch == 'r':
                    if oldimage:
                        ptty.driver.reset()
                        ptty.driver.init(partial=False)
                        ptty.driver.draw(0, 0, oldimage)
                        ptty.driver.reset()
                        ptty.driver.init(ptty.partial)

            # if user or SIGUSR1 toggled the scrub flag, scrub display and start with a fresh image
            if flags['scrub_requested']:
                ptty.driver.scrub()
                # clear old image and buffer and restore flag
                oldimage = None
                oldbuff = ''
                flags['scrub_requested'] = False
            
            with open(vcsa, 'rb') as f:
                with open(vcsudev, 'rb') as vcsu:
                    # read the first 4 bytes to get the console attributes
                    attributes = f.read(4)
                    rows, cols, x, y = list(map(ord, struct.unpack('cccc', attributes)))

                    # read from the text buffer 
                    buff = vcsu.read()
                    if character_width == 4:
                        # work around weird bug
                        buff = buff.replace(b'\x20\x20\x20\x20', b'\x20\x00\x00\x00')
                    # find character under cursor (in case using a non-fixed width font)
                    char_under_cursor = buff[character_width * (y * rows + x):character_width * (y * rows + x + 1)]
                    encoding = 'utf_32' if character_width == 4 else ptty.encoding
                    cursor = (x, y, char_under_cursor.decode(encoding, 'ignore'))
                    # add newlines per column count
                    buff = ''.join([r.decode(encoding, 'replace') + '\n' for r in ptty.split(buff, cols * character_width)])
                    # do something only if content has changed or cursor was moved
                    if buff != oldbuff or cursor != oldcursor:
                        # show new content
                        oldimage = ptty.showtext(buff, fill=ptty.black, cursor=cursor if not nocursor else None,
                                                oldimage=oldimage,
                                                **textargs)
                        oldbuff = buff
                        oldcursor = cursor
                    else:
                        # delay before next update check
                        time.sleep(float(sleep))


# add all the CLI commands
cli.add_command(scrub)
cli.add_command(terminal)
cli.add_command(stdin)
cli.add_command(image)
cli.add_command(vnc)
cli.add_command(fb)
cli.add_command(list_drivers)


if __name__ == '__main__':
    cli()



#     Copyright (c) 2018 Jouko Strömmer
#     Copyright (c) 2017 Waveshare
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from papertty.drivers.drivers_full import WaveshareFull

from PIL import Image



class WaveshareColor(WaveshareFull):
    """Base class for 'color' displays, the B/C variants: black-white-red and black-white-yellow. This includes:
    - 4.2" B (uses two separate frame buffers - one for B/W and one for red)
    - 7.5" B (uses one frame buffer - black < 64 < red < 192 < white)
    """

    VCM_DC_SETTING = 0x82

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.colors = 3

    def display_frame(self, frame_buffer, *args):
        pass

    def init(self, **kwargs):
        pass

    def draw(self, x, y, image):
        """Display an image - this module does not support partial refresh: x, y are ignored
        IMPORTANT NOTE: this method IGNORES the red buffer completely!"""
        self.display_frame(self.get_frame_buffer(image))


class EPD4in2b(WaveshareColor):
    """Waveshare 4.2" B - black / white / red"""

    ACTIVE_PROGRAM = 0xA1
    B2B_LUT = 0x24
    B2W_LUT = 0x22
    DATA_START_TRANSMISSION_2 = 0x13
    GSST_SETTING = 0x65
    PARTIAL_IN = 0x91
    PARTIAL_OUT = 0x92
    PARTIAL_WINDOW = 0x90
    POWER_SAVING = 0xE3
    PROGRAM_MODE = 0xA0
    READ_OTP_DATA = 0xA2
    RESOLUTION_SETTING = 0x61
    TEMPERATURE_SENSOR_CALIBRATION = 0x40
    TEMPERATURE_SENSOR_SELECTION = 0x41
    VCOM_LUT = 0x20
    VCOM_VALUE = 0x81
    W2B_LUT = 0x23
    W2W_LUT = 0x21

    def __init__(self):
        super().__init__(name='4.2" B', width=400, height=300)

    def init(self, **kwargs):
        if self.epd_init() != 0:
            return -1
        self.reset()
        self.send_command(self.BOOSTER_SOFT_START)
        self.send_data(0x17)
        self.send_data(0x17)
        self.send_data(0x17)  # 07 0f 17 1f 27 2F 37 2f
        self.send_command(self.POWER_ON)
        self.wait_until_idle()
        self.send_command(self.PANEL_SETTING)
        self.send_data(0x0F)  # LUT from OTP

    def get_frame_buffer(self, image, reverse=True):
        super().get_frame_buffer(image, reverse=reverse)

    def display_frame(self, frame_buffer_black, *args):
        frame_buffer_red = args[0] if args else None
        if frame_buffer_black:
            self.send_command(self.DATA_START_TRANSMISSION_1)
            self.delay_ms(2)
            for i in range(0, int(self.width * self.height / 8)):
                self.send_data(frame_buffer_black[i])
            self.delay_ms(2)
        if frame_buffer_red:
            self.send_command(self.DATA_START_TRANSMISSION_2)
            self.delay_ms(2)
            for i in range(0, int(self.width * self.height / 8)):
                self.send_data(frame_buffer_red[i])
            self.delay_ms(2)

        self.send_command(self.DISPLAY_REFRESH)
        self.wait_until_idle()

    # after this, call epd.init() to awaken the module
    def sleep(self):
        self.send_command(self.VCOM_AND_DATA_INTERVAL_SETTING)
        self.send_data(0xF7)  # border floating
        self.send_command(self.POWER_OFF)
        self.wait_until_idle()
        self.send_command(self.DEEP_SLEEP)
        self.send_data(0xA5)  # check code


class EPD7in5b(WaveshareColor):
    """Waveshare 7.5" B - black / white / red"""

    IMAGE_PROCESS = 0x13
    LUT_BLUE = 0x21
    LUT_GRAY_1 = 0x23
    LUT_GRAY_2 = 0x24
    LUT_RED_0 = 0x25
    LUT_RED_1 = 0x26
    LUT_RED_2 = 0x27
    LUT_RED_3 = 0x28
    LUT_WHITE = 0x22
    LUT_XON = 0x29
    READ_VCOM_VALUE = 0x81
    REVISION = 0x70
    SPI_FLASH_CONTROL = 0x65
    TCON_RESOLUTION = 0x61
    TEMPERATURE_CALIBRATION = 0x41

    def __init__(self):
        super().__init__(name='7.5" B', width=640, height=384)

    def init(self, **kwargs):
        if self.epd_init() != 0:
            return -1
        self.reset()
        self.send_command(self.POWER_SETTING)
        self.send_data(0x37)
        self.send_data(0x00)
        self.send_command(self.PANEL_SETTING)
        self.send_data(0xCF)
        self.send_data(0x08)
        self.send_command(self.BOOSTER_SOFT_START)
        self.send_data(0xc7)
        self.send_data(0xcc)
        self.send_data(0x28)
        self.send_command(self.POWER_ON)
        self.wait_until_idle()
        self.send_command(self.PLL_CONTROL)
        self.send_data(0x3c)
        self.send_command(self.TEMPERATURE_CALIBRATION)
        self.send_data(0x00)
        self.send_command(self.VCOM_AND_DATA_INTERVAL_SETTING)
        self.send_data(0x77)
        self.send_command(self.TCON_SETTING)
        self.send_data(0x22)
        self.send_command(self.TCON_RESOLUTION)
        self.send_data(0x02)  # source 640
        self.send_data(0x80)
        self.send_data(0x01)  # gate 384
        self.send_data(0x80)
        self.send_command(self.VCM_DC_SETTING)
        self.send_data(0x1E)  # decide by LUT file
        self.send_command(0xe5)  # FLASH MODE
        self.send_data(0x03)

    def get_frame_buffer(self, image, reverse=False):
        buf = [0x00] * int(self.width * self.height / 4)
        # Set buffer to value of Python Imaging Library image.
        # Image must be in mode L.
        image_grayscale = image.convert('L')
        imwidth, imheight = image_grayscale.size
        if imwidth != self.width or imheight != self.height:
            raise ValueError('Image must be same dimensions as display \
                ({0}x{1}).'.format(self.width, self.height))

        pixels = image_grayscale.load()
        for y in range(self.height):
            for x in range(self.width):
                # Set the bits for the column of pixels at the current position.
                if pixels[x, y] < 64:  # black
                    buf[int((x + y * self.width) / 4)] &= ~(0xC0 >> (x % 4 * 2))
                elif pixels[x, y] < 192:  # convert gray to red
                    buf[int((x + y * self.width) / 4)] &= ~(0xC0 >> (x % 4 * 2))
                    buf[int((x + y * self.width) / 4)] |= 0x40 >> (x % 4 * 2)
                else:  # white
                    buf[int((x + y * self.width) / 4)] |= 0xC0 >> (x % 4 * 2)
        return buf

    def display_frame(self, frame_buffer, *args):
        self.send_command(self.DATA_START_TRANSMISSION_1)
        for i in range(0, int(self.width / 4 * self.height)):
            temp1 = frame_buffer[i]
            j = 0
            while j < 4:
                if (temp1 & 0xC0) == 0xC0:
                    temp2 = 0x03
                elif (temp1 & 0xC0) == 0x00:
                    temp2 = 0x00
                else:
                    temp2 = 0x04
                temp2 = (temp2 << 4) & 0xFF
                temp1 = (temp1 << 2) & 0xFF
                j += 1
                if (temp1 & 0xC0) == 0xC0:
                    temp2 |= 0x03
                elif (temp1 & 0xC0) == 0x00:
                    temp2 |= 0x00
                else:
                    temp2 |= 0x04
                temp1 = (temp1 << 2) & 0xFF
                self.send_data(temp2)
                j += 1
        self.send_command(self.DISPLAY_REFRESH)
        self.delay_ms(100)
        self.wait_until_idle()

    def sleep(self):
        self.send_command(self.POWER_OFF)
        self.wait_until_idle()
        self.send_command(self.DEEP_SLEEP)
        self.send_data(0xa5)


class EPD5in65f(WaveshareColor):
    """Waveshare 5.65" - 7 colors"""

    LUT_BLACK = 0x000000  # 0000  BGR
    LUT_WHITE = 0xffffff  # 0001
    LUT_GREEN = 0x00ff00  # 0010
    LUT_BLUE = 0xff0000  # 0011
    LUT_RED = 0x0000ff  # 0100
    LUT_YELLOW = 0x00ffff  # 0101
    LUT_ORANGE = 0x0080ff  # 0110
    POWER_SAVING = 0xE3
    TCON_RESOLUTION = 0x61
    TEMPERATURE_CALIBRATION = 0x41

    def __init__(self):
        super().__init__(name='5.65" F', width=600, height=448)

    def reset(self):
        self.digital_write(self.RST_PIN, 1)
        self.delay_ms(600)
        self.digital_write(self.RST_PIN, 0)
        self.delay_ms(2)
        self.digital_write(self.RST_PIN, 1)
        self.delay_ms(200)

    def send_command(self, command):
        self.digital_write(self.DC_PIN, 0)
        self.digital_write(self.CS_PIN, 0)
        self.spi_transfer([command])
        self.digital_write(self.CS_PIN, 1)

    def send_data(self, data):
        self.digital_write(self.DC_PIN, 1)
        self.digital_write(self.CS_PIN, 0)
        self.spi_transfer([data])
        self.digital_write(self.CS_PIN, 1)

    def wait_until_busy(self):
        while self.digital_read(self.BUSY_PIN) == 0:  # 0: idle, 1: busy
            self.delay_ms(100)

    def wait_until_idle(self):
        while self.digital_read(self.BUSY_PIN) == 1:  # 0: idle, 1: busy
            self.delay_ms(100)

    def init(self, **kwargs):
        if self.epd_init() != 0:
            return -1

        # EPD hardware init start
        self.reset()

        self.wait_until_busy()
        self.send_command(self.PANEL_SETTING)
        self.send_data(0xEF)
        self.send_data(0x08)
        self.send_command(self.POWER_SETTING)
        self.send_data(0x37)
        self.send_data(0x00)
        self.send_data(0x23)
        self.send_data(0x23)
        self.send_command(self.POWER_OFF_SEQUENCE_SETTING)
        self.send_data(0x00)
        self.send_command(self.BOOSTER_SOFT_START)
        self.send_data(0xC7)
        self.send_data(0xC7)
        self.send_data(0x1D)
        self.send_command(self.PLL_CONTROL)
        self.send_data(0x3C)
        self.send_command(self.TEMPERATURE_SENSOR_COMMAND)
        self.send_data(0x00)
        self.send_command(self.VCOM_AND_DATA_INTERVAL_SETTING)
        self.send_data(0x37)
        self.send_command(self.TCON_SETTING)
        self.send_data(0x22)
        self.send_command(self.TCON_RESOLUTION)
        self.send_data(0x02)
        self.send_data(0x58)
        self.send_data(0x01)
        self.send_data(0xC0)
        self.send_command(self.POWER_SAVING)
        self.send_data(0xAA)

        self.delay_ms(100)
        self.send_command(0x50)
        self.send_data(0x37)

    def get_frame_buffer(self, image, reverse=False):
        buf = [0x00] * int(self.width * self.height / 2)

        # quantize image with palette of possible colors
        image_palette = Image.new('P', (1, 1))
        image_palette.putpalette(([0, 0, 0, 255, 255, 255, 0, 255, 0, 0, 0, 255, 255, 0, 0, 255, 255, 0, 255, 128, 0]
                                  + [0, 0, 0]) * 32) # multiply by 32 to pad palette to reach required length of 768
        image_rgb = image.quantize(palette=image_palette).convert('RGB')

        imwidth, imheight = image_rgb.size

        if imwidth != self.width or imheight != self.height:
            raise ValueError('Image must be same dimensions as display \
                ({0}x{1}).'.format(self.width, self.height))

        pixels = image_rgb.load()

        for y in range(self.height):
            for x in range(self.width):
                pos = int((x + y * self.width) / 2)
                color = 0

                if pixels[x, y][0] == 0 and pixels[x, y][1] == 0 and pixels[x, y][2] == 0:
                    color = 0
                elif pixels[x, y][0] == 255 and pixels[x, y][1] == 255 and pixels[x, y][2] == 255:
                    color = 1
                elif pixels[x, y][0] == 0 and pixels[x, y][1] == 255 and pixels[x, y][2] == 0:
                    color = 2
                elif pixels[x, y][0] == 0 and pixels[x, y][1] == 0 and pixels[x, y][2] == 255:
                    color = 3
                elif pixels[x, y][0] == 255 and pixels[x, y][1] == 0 and pixels[x, y][2] == 0:
                    color = 4
                elif pixels[x, y][0] == 255 and pixels[x, y][1] == 255 and pixels[x, y][2] == 0:
                    color = 5
                elif pixels[x, y][0] == 255 and pixels[x, y][1] == 128 and pixels[x, y][2] == 0:
                    color = 6

                data_t = buf[pos] & (~(0xF0 >> ((x % 2) * 4)))
                buf[pos] = data_t | ((color << 4) >> ((x % 2) * 4))

        return buf

    def display_frame(self, frame_buffer, *args):
        self.send_command(self.TCON_RESOLUTION)
        self.send_data(0x02)
        self.send_data(0x58)
        self.send_data(0x01)
        self.send_data(0xC0)
        self.send_command(self.DATA_START_TRANSMISSION_1)
        for i in range(0, int(self.height)):
            for j in range(0, int(self.width / 2)):
                self.send_data((frame_buffer[j + (int(self.width / 2) * i)]))
        self.send_command(self.POWER_ON)
        self.wait_until_busy()
        self.send_command(self.DISPLAY_REFRESH)
        self.wait_until_busy()
        self.send_command(self.POWER_OFF)
        self.wait_until_idle()
        self.delay_ms(500)

    def sleep(self):
        self.delay_ms(500)
        self.send_command(self.DEEP_SLEEP)
        self.send_data(0XA5)
        self.digital_write(self.RST_PIN, 0)


# This is a 'monochrome' display but surprisingly, the only difference to EPD7in5b is
# setting a different resolution in init() (after TCON_RESOLUTION command),
# thus, it is here and a subclass of EPD7in5b (for now).
class EPD5in83(EPD7in5b):
    """Waveshare 5.83" - monochrome"""

    def __init__(self, **kwargs):
        super().__init__()
        self.name = '5.83" BW'
        self.width = 600
        self.height = 448

    def init(self, **kwargs):
        if self.epd_init() != 0:
            return -1
        self.reset()
        self.send_command(self.POWER_SETTING)
        self.send_data(0x37)
        self.send_data(0x00)
        self.send_command(self.PANEL_SETTING)
        self.send_data(0xCF)
        self.send_data(0x08)
        self.send_command(self.BOOSTER_SOFT_START)
        self.send_data(0xc7)
        self.send_data(0xcc)
        self.send_data(0x28)
        self.send_command(self.POWER_ON)
        self.wait_until_idle()
        self.send_command(self.PLL_CONTROL)
        self.send_data(0x3c)
        self.send_command(self.TEMPERATURE_CALIBRATION)
        self.send_data(0x00)
        self.send_command(self.VCOM_AND_DATA_INTERVAL_SETTING)
        self.send_data(0x77)
        self.send_command(self.TCON_SETTING)
        self.send_data(0x22)
        self.send_command(self.TCON_RESOLUTION)
        self.send_data(0x02)  # source 600
        self.send_data(0x58)
        self.send_data(0x01)  # gate 448
        self.send_data(0xC0)
        self.send_command(self.VCM_DC_SETTING)
        self.send_data(0x1E)  # decide by LUT file
        self.send_command(0xe5)  # FLASH MODE
        self.send_data(0x03)


class EPD5in83b(EPD5in83):
    """Waveshare 5.83" B - black / white / red"""

    def __init__(self, **kwargs):
        super().__init__()
        self.name = '5.83" B'

    def init(self, **kwargs):
        if self.epd_init() != 0:
            return -1
        self.reset()
        self.send_command(self.POWER_SETTING)
        self.send_data(0x37)
        self.send_data(0x00)
        self.send_command(self.PANEL_SETTING)
        self.send_data(0xCF)
        self.send_data(0x08)
        self.send_command(self.BOOSTER_SOFT_START)
        self.send_data(0xc7)
        self.send_data(0xcc)
        self.send_data(0x28)
        self.send_command(self.POWER_ON)
        self.wait_until_idle()
        self.send_command(self.PLL_CONTROL)
        self.send_data(0x3c)
        self.send_command(self.TEMPERATURE_CALIBRATION)
        self.send_data(0x00)
        self.send_command(self.VCOM_AND_DATA_INTERVAL_SETTING)
        self.send_data(0x77)
        self.send_command(self.TCON_SETTING)
        self.send_data(0x22)
        self.send_command(self.TCON_RESOLUTION)
        self.send_data(0x02)  # source 600
        self.send_data(0x58)
        self.send_data(0x01)  # gate 448
        self.send_data(0xc0)
        self.send_command(self.VCM_DC_SETTING)
        self.send_data(0x20)  # decide by LUT file
        self.send_command(0xe5)  # FLASH MODE
        self.send_data(0x03)

    def get_frame_buffer(self, image, reverse=True):
        super().get_frame_buffer(image, reverse=reverse)

    def display_frame(self, frame_buffer_black, *args):
        frame_buffer_red = args[0] if args else None
        self.send_command(self.DATA_START_TRANSMISSION_1)
        for i in range(0, int(self.width / 8 * self.height)):
            temp1 = frame_buffer_black[i]
            temp2 = frame_buffer_red[i]
            j = 0
            while j < 8:
                if (temp2 & 0x80) == 0x00:
                    temp3 = 0x04  # red
                elif (temp1 & 0x80) == 0x00:
                    temp3 = 0x00  # black
                else:
                    temp3 = 0x03  # white

                temp3 = (temp3 << 4) & 0xFF
                temp1 = (temp1 << 1) & 0xFF
                temp2 = (temp2 << 1) & 0xFF
                j += 1
                if (temp2 & 0x80) == 0x00:
                    temp3 |= 0x04  # red
                elif (temp1 & 0x80) == 0x00:
                    temp3 |= 0x00  # black
                else:
                    temp3 |= 0x03  # white
                temp1 = (temp1 << 1) & 0xFF
                temp2 = (temp2 << 1) & 0xFF
                self.send_data(temp3)
                j += 1
        self.send_command(self.DISPLAY_REFRESH)
        self.delay_ms(100)
        self.wait_until_idle()



#     Copyright (c) 2018 Jouko Strömmer
#     Copyright (c) 2017 Waveshare
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with this program.  If not, see <https://www.gnu.org/licenses/>.

from PIL import Image, ImageDraw

from papertty.drivers.drivers_color import WaveshareColor


class WaveshareColorDraw(WaveshareColor):
    """Class for displays that have (mostly shared) special implementations of some drawing methods. This includes:
    - 1.54" B
    - 1.54" C
    - 2.13" B
    - 2.7" B
    - 2.9" B
    
    """

    ACTIVE_PROGRAM = 0xA1
    DATA_START_TRANSMISSION_2 = 0x13
    PROGRAM_MODE = 0xA0
    READ_OTP_DATA = 0xA2
    VCOM_VALUE = 0x81

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.colors = 3
        self.rotate = self.ROTATE_0
        # the rotation aware algorithms refer to and swap the screen dimensions, so set them here
        self.EPD_WIDTH = kwargs['width']
        self.EPD_HEIGHT = kwargs['height']

    # EPD1in54b and EPD1in54c use 0x17
    # EPD2in13b and EPD2in9b use 0x37
    # EPD2in7b has a simpler sleep
    def sleep(self, sleepbyte=0x17):
        self.send_command(self.VCOM_AND_DATA_INTERVAL_SETTING)
        self.send_data(sleepbyte)
        self.send_command(self.VCM_DC_SETTING)  # to solve Vcom drop
        self.send_data(0x00)
        self.send_command(self.POWER_SETTING)  # power setting
        self.send_data(0x02)  # gate switch to external
        self.send_data(0x00)
        self.send_data(0x00)
        self.send_data(0x00)
        self.wait_until_idle()
        self.send_command(self.POWER_OFF)  # power off

    def set_rotate(self, rotate):
        if rotate == self.ROTATE_0:
            self.rotate = self.ROTATE_0
            self.width = self.EPD_WIDTH
            self.height = self.EPD_HEIGHT
        elif rotate == self.ROTATE_90:
            self.rotate = self.ROTATE_90
            self.width = self.EPD_HEIGHT
            self.height = self.EPD_WIDTH
        elif rotate == self.ROTATE_180:
            self.rotate = self.ROTATE_180
            self.width = self.EPD_WIDTH
            self.height = self.EPD_HEIGHT
        elif rotate == self.ROTATE_270:
            self.rotate = self.ROTATE_270
            self.width = self.EPD_HEIGHT
            self.height = self.EPD_WIDTH

    # this variant is for EPD1in54c, EPD2in13b and EPD2in9b - EPD1in54b and EPD2in7b override it
    def display_frame(self, frame_buffer_black, *args):
        frame_buffer_red = args[0] if args else None
        if frame_buffer_black:
            self.send_command(self.DATA_START_TRANSMISSION_1)
            self.delay_ms(2)
            for i in range(0, int(self.width * self.height / 8)):
                self.send_data(frame_buffer_black[i])
            self.delay_ms(2)
        if frame_buffer_red:
            self.send_command(self.DATA_START_TRANSMISSION_2)
            self.delay_ms(2)
            for i in range(0, int(self.width * self.height / 8)):
                self.send_data(frame_buffer_red[i])
            self.delay_ms(2)

        self.send_command(self.DISPLAY_REFRESH)
        self.wait_until_idle()

    # for EPD154* - override elsewhere
    def set_lut_bw(self):
        self.send_command(0x20)  # vcom
        for count in range(0, 15):
            self.send_data(self.lut_vcom0[count])
        self.send_command(0x21)  # ww --
        for count in range(0, 15):
            self.send_data(self.lut_w[count])
        self.send_command(0x22)  # bw r
        for count in range(0, 15):
            self.send_data(self.lut_b[count])
        self.send_command(0x23)  # wb w
        for count in range(0, 15):
            self.send_data(self.lut_g1[count])
        self.send_command(0x24)  # bb b
        for count in range(0, 15):
            self.send_data(self.lut_g2[count])

    # for EPD154* - override elsewhere
    def set_lut_red(self):
        self.send_command(0x25)
        for count in range(0, 15):
            self.send_data(self.lut_vcom1[count])
        self.send_command(0x26)
        for count in range(0, 15):
            self.send_data(self.lut_red0[count])
        self.send_command(0x27)
        for count in range(0, 15):
            self.send_data(self.lut_red1[count])

    # these LUTs are for EPD154* - override in 2.7"
    lut_vcom0 = [
        0x0E, 0x14, 0x01, 0x0A, 0x06, 0x04, 0x0A, 0x0A,
        0x0F, 0x03, 0x03, 0x0C, 0x06, 0x0A, 0x00
    ]

    lut_w = [
        0x0E, 0x14, 0x01, 0x0A, 0x46, 0x04, 0x8A, 0x4A,
        0x0F, 0x83, 0x43, 0x0C, 0x86, 0x0A, 0x04
    ]

    lut_b = [
        0x0E, 0x14, 0x01, 0x8A, 0x06, 0x04, 0x8A, 0x4A,
        0x0F, 0x83, 0x43, 0x0C, 0x06, 0x4A, 0x04
    ]

    lut_g1 = [
        0x8E, 0x94, 0x01, 0x8A, 0x06, 0x04, 0x8A, 0x4A,
        0x0F, 0x83, 0x43, 0x0C, 0x06, 0x0A, 0x04
    ]

    lut_g2 = [
        0x8E, 0x94, 0x01, 0x8A, 0x06, 0x04, 0x8A, 0x4A,
        0x0F, 0x83, 0x43, 0x0C, 0x06, 0x0A, 0x04
    ]

    lut_vcom1 = [
        0x03, 0x1D, 0x01, 0x01, 0x08, 0x23, 0x37, 0x37,
        0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    ]

    lut_red0 = [
        0x83, 0x5D, 0x01, 0x81, 0x48, 0x23, 0x77, 0x77,
        0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    ]

    lut_red1 = [
        0x03, 0x1D, 0x01, 0x01, 0x08, 0x23, 0x37, 0x37,
        0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
    ]

    def set_pixel(self, frame_buffer, x, y, colored):
        if x < 0 or x >= self.width or y < 0 or y >= self.height:
            return
        if self.rotate == self.ROTATE_0:
            self.set_absolute_pixel(frame_buffer, x, y, colored)
        elif self.rotate == self.ROTATE_90:
            point_temp = x
            x = self.EPD_WIDTH - y
            y = point_temp
            self.set_absolute_pixel(frame_buffer, x, y, colored)
        elif self.rotate == self.ROTATE_180:
            x = self.EPD_WIDTH - x
            y = self.EPD_HEIGHT - y
            self.set_absolute_pixel(frame_buffer, x, y, colored)
        elif self.rotate == self.ROTATE_270:
            point_temp = x
            x = y
            y = self.EPD_HEIGHT - point_temp
            self.set_absolute_pixel(frame_buffer, x, y, colored)

    # overridden in EPD2in7b
    def set_absolute_pixel(self, frame_buffer, x, y, colored, reverse=False):
        # To avoid display orientation effects
        # use EPD_WIDTH instead of self.width
        # use EPD_HEIGHT instead of self.height
        if x < 0 or x >= self.EPD_WIDTH or y < 0 or y >= self.EPD_HEIGHT:
            return
        if not colored if reverse else colored:
            frame_buffer[(x + y * self.EPD_WIDTH) / 8] &= ~(0x80 >> (x % 8))
        else:
            frame_buffer[(x + y * self.EPD_WIDTH) / 8] |= 0x80 >> (x % 8)

    def draw_circle(self, frame_buffer, x, y, radius, colored):
        # Bresenham algorithm
        x_pos = -radius
        y_pos = 0
        err = 2 - 2 * radius
        if x >= self.width or y >= self.height:
            return
        while True:
            self.set_pixel(frame_buffer, x - x_pos, y + y_pos, colored)
            self.set_pixel(frame_buffer, x + x_pos, y + y_pos, colored)
            self.set_pixel(frame_buffer, x + x_pos, y - y_pos, colored)
            self.set_pixel(frame_buffer, x - x_pos, y - y_pos, colored)
            e2 = err
            if e2 <= y_pos:
                y_pos += 1
                err += y_pos * 2 + 1
                if -x_pos == y_pos and e2 <= x_pos:
                    e2 = 0
            if e2 > x_pos:
                x_pos += 1
                err += x_pos * 2 + 1
            if x_pos > 0:
                break

    # this only appears in EPD1in54b and EPD1in54c source
    def display_string_at(self, frame_buffer, x, y, text, font, colored):
        image = Image.new('1', (self.width, self.height))
        draw = ImageDraw.Draw(image)
        draw.text((x, y), text, font=font, fill=255)
        # Set buffer to value of Python Imaging Library image.
        # Image must be in mode 1.
        pixels = image.load()
        for y in range(self.height):
            for x in range(self.width):
                # Set the bits for the column of pixels at the current position.
                if pixels[x, y] != 0:
                    self.set_pixel(frame_buffer, x, y, colored)

    # this, on the other hand, appears in the EPD2in7b source - same method, different name
    def draw_string_at(self, frame_buffer, x, y, text, font, colored):
        self.display_string_at(frame_buffer, x, y, text, font, colored)

    def draw_line(self, frame_buffer, x0, y0, x1, y1, colored):
        # Bresenham algorithm
        dx = abs(x1 - x0)
        sx = 1 if x0 < x1 else -1
        dy = -abs(y1 - y0)
        sy = 1 if y0 < y1 else -1
        err = dx + dy
        while (x0 != x1) and (y0 != y1):
            self.set_pixel(frame_buffer, x0, y0, colored)
            if 2 * err >= dy:
                err += dy
                x0 += sx
            if 2 * err <= dx:
                err += dx
                y0 += sy

    def draw_horizontal_line(self, frame_buffer, x, y, width, colored):
        for i in range(x, x + width):
            self.set_pixel(frame_buffer, i, y, colored)

    def draw_vertical_line(self, frame_buffer, x, y, height, colored):
        for i in range(y, y + height):
            self.set_pixel(frame_buffer, x, i, colored)

    def draw_rectangle(self, frame_buffer, x0, y0, x1, y1, colored):
        min_x = x0 if x1 > x0 else x1
        max_x = x1 if x1 > x0 else x0
        min_y = y0 if y1 > y0 else y1
        max_y = y1 if y1 > y0 else y0
        self.draw_horizontal_line(frame_buffer, min_x, min_y, max_x - min_x + 1, colored)
        self.draw_horizontal_line(frame_buffer, min_x, max_y, max_x - min_x + 1, colored)
        self.draw_vertical_line(frame_buffer, min_x, min_y, max_y - min_y + 1, colored)
        self.draw_vertical_line(frame_buffer, max_x, min_y, max_y - min_y + 1, colored)

    def draw_filled_rectangle(self, frame_buffer, x0, y0, x1, y1, colored):
        min_x = x0 if x1 > x0 else x1
        max_x = x1 if x1 > x0 else x0
        min_y = y0 if y1 > y0 else y1
        max_y = y1 if y1 > y0 else y0
        for i in range(min_x, max_x + 1):
            self.draw_vertical_line(frame_buffer, i, min_y, max_y - min_y + 1, colored)

    def draw_filled_circle(self, frame_buffer, x, y, radius, colored):
        # Bresenham algorithm
        x_pos = -radius
        y_pos = 0
        err = 2 - 2 * radius
        if x >= self.width or y >= self.height:
            return
        while True:
            self.set_pixel(frame_buffer, x - x_pos, y + y_pos, colored)
            self.set_pixel(frame_buffer, x + x_pos, y + y_pos, colored)
            self.set_pixel(frame_buffer, x + x_pos, y - y_pos, colored)
            self.set_pixel(frame_buffer, x - x_pos, y - y_pos, colored)
            self.draw_horizontal_line(frame_buffer, x + x_pos, y + y_pos, 2 * (-x_pos) + 1, colored)
            self.draw_horizontal_line(frame_buffer, x + x_pos, y - y_pos, 2 * (-x_pos) + 1, colored)
            e2 = err
            if e2 <= y_pos:
                y_pos += 1
                err += y_pos * 2 + 1
                if -x_pos == y_pos and e2 <= x_pos:
                    e2 = 0
            if e2 > x_pos:
                x_pos += 1
                err += x_pos * 2 + 1
            if x_pos > 0:
                break


class EPD1in54b(WaveshareColorDraw):
    """Waveshare 1.54" B - black / white / red"""

    SOURCE_AND_GATE_START_SETTING = 0x62
    TCON_RESOLUTION = 0x61
    TEMPERATURE_SENSOR_CALIBRATION = 0x41

    def __init__(self):
        super().__init__(name='1.54" B', width=200, height=200)

    def init(self, **kwargs):
        if self.epd_init() != 0:
            return -1
        self.reset()
        self.send_command(self.POWER_SETTING)
        self.send_data(0x07)
        self.send_data(0x00)
        self.send_data(0x08)
        self.send_data(0x00)
        self.send_command(self.BOOSTER_SOFT_START)
        self.send_data(0x07)
        self.send_data(0x07)
        self.send_data(0x07)
        self.send_command(self.POWER_ON)

        self.wait_until_idle()

        self.send_command(self.PANEL_SETTING)
        self.send_data(0xCF)
        self.send_command(self.VCOM_AND_DATA_INTERVAL_SETTING)
        self.send_data(0x17)
        self.send_command(self.PLL_CONTROL)
        self.send_data(0x39)
        self.send_command(self.TCON_RESOLUTION)
        self.send_data(0xC8)
        self.send_data(0x00)
        self.send_data(0xC8)
        self.send_command(self.VCM_DC_SETTING)
        self.send_data(0x0E)

        self.set_lut_bw()
        self.set_lut_red()
        return 0

    def get_frame_buffer(self, image, reverse=True):
        super().get_frame_buffer(image, reverse=reverse)

    def display_frame(self, frame_buffer_black, *args):
        frame_buffer_red = args[0] if args else None
        if frame_buffer_black:
            self.send_command(self.DATA_START_TRANSMISSION_1)
            self.delay_ms(2)
            for i in range(0, int(self.width * self.height / 8)):
                temp = 0x00
                for bit in range(0, 4):
                    if frame_buffer_black[i] & (0x80 >> bit) != 0:
                        temp |= 0xC0 >> (bit * 2)
                self.send_data(temp)
                temp = 0x00
                for bit in range(4, 8):
                    if frame_buffer_black[i] & (0x80 >> bit) != 0:
                        temp |= 0xC0 >> ((bit - 4) * 2)
                self.send_data(temp)
            self.delay_ms(2)
        if frame_buffer_red:
            self.send_command(self.DATA_START_TRANSMISSION_2)
            self.delay_ms(2)
            for i in range(0, int(self.width * self.height / 8)):
                self.send_data(frame_buffer_red[i])
            self.delay_ms(2)

        self.send_command(self.DISPLAY_REFRESH)
        self.wait_until_idle()


class EPD1in54c(WaveshareColorDraw):
    """Waveshare 1.54" C - black / white / yellow"""

    SOURCE_AND_GATE_START_SETTING = 0x62
    TCON_RESOLUTION = 0x61
    TEMPERATURE_SENSOR_CALIBRATION = 0x41

    def __init__(self):
        super().__init__(name='1.54" C', width=152, height=152)

    def init(self, **kwargs):
        if self.epd_init() != 0:
            return -1
        self.reset()
        self.send_command(self.POWER_SETTING)
        self.send_data(0x07)
        self.send_data(0x00)
        self.send_data(0x08)
        self.send_data(0x00)
        self.send_command(self.BOOSTER_SOFT_START)
        self.send_data(0x17)
        self.send_data(0x17)
        self.send_data(0x17)
        self.send_command(self.POWER_ON)

        self.wait_until_idle()

        self.send_command(self.PANEL_SETTING)
        self.send_data(0x0F)
        self.send_data(0x0D)
        self.send_command(self.VCOM_AND_DATA_INTERVAL_SETTING)
        self.send_data(0x77)
        self.send_command(self.PLL_CONTROL)
        self.send_data(0x39)
        self.send_command(self.TCON_RESOLUTION)
        self.send_data(0x98)
        self.send_data(0x00)
        self.send_data(0x98)
        self.send_command(self.VCM_DC_SETTING)
        self.send_data(0x0E)

        self.set_lut_bw()
        self.set_lut_red()
        return 0

    def get_frame_buffer(self, image, reverse=True):
        super().get_frame_buffer(image, reverse=reverse)


class EPD2in13b(WaveshareColorDraw):
    """Waveshare 2.13" B - black / white / red"""

    B2B_LUT = 0x24
    B2W_LUT = 0x22
    PARTIAL_IN = 0x91
    PARTIAL_OUT = 0x92
    PARTIAL_WINDOW = 0x90
    POWER_SAVING = 0xE3
    RESOLUTION_SETTING = 0x61
    TEMPERATURE_SENSOR_CALIBRATION = 0x40
    TEMPERATURE_SENSOR_SELECTION = 0x41
    VCOM_LUT = 0x20
    W2B_LUT = 0x23
    W2W_LUT = 0x21

    def __init__(self):
        super().__init__(name='2.13" B', width=104, height=212)

    def init(self, **kwargs):
        if self.epd_init() != 0:
            return -1
        self.reset()
        self.send_command(self.BOOSTER_SOFT_START)
        self.send_data(0x17)
        self.send_data(0x17)
        self.send_data(0x17)
        self.send_command(self.POWER_ON)
        self.wait_until_idle()
        self.send_command(self.PANEL_SETTING)
        self.send_data(0x8F)
        self.send_command(self.VCOM_AND_DATA_INTERVAL_SETTING)
        self.send_data(0x37)
        self.send_command(self.RESOLUTION_SETTING)
        self.send_data(0x68)
        self.send_data(0x00)
        self.send_data(0xD4)

    def get_frame_buffer(self, image, reverse=True):
        super().get_frame_buffer(image, reverse=reverse)

    def sleep(self, sleepbyte=0x37):
        super().sleep(sleepbyte=sleepbyte)


class EPD2in7b(WaveshareColorDraw):
    """Waveshare 2.7" B - black / white / red"""

    LUT_BLACK_TO_BLACK = 0x24
    LUT_BLACK_TO_WHITE = 0x22
    LUT_WHITE_TO_BLACK = 0x23
    LUT_WHITE_TO_WHITE = 0x21
    PARTIAL_DATA_START_TRANSMISSION_1 = 0x14
    PARTIAL_DATA_START_TRANSMISSION_2 = 0x15
    PARTIAL_DISPLAY_REFRESH = 0x16
    SOURCE_AND_GATE_START_SETTING = 0x62
    TCON_RESOLUTION = 0x61
    TEMPERATURE_SENSOR_CALIBRATION = 0x41

    lut_vcom_dc = [
        0x00, 0x00,
        0x00, 0x1A, 0x1A, 0x00, 0x00, 0x01,
        0x00, 0x0A, 0x0A, 0x00, 0x00, 0x08,
        0x00, 0x0E, 0x01, 0x0E, 0x01, 0x10,
        0x00, 0x0A, 0x0A, 0x00, 0x00, 0x08,
        0x00, 0x04, 0x10, 0x00, 0x00, 0x05,
        0x00, 0x03, 0x0E, 0x00, 0x00, 0x0A,
        0x00, 0x23, 0x00, 0x00, 0x00, 0x01
    ]

    # R21H
    lut_ww = [
        0x90, 0x1A, 0x1A, 0x00, 0x00, 0x01,
        0x40, 0x0A, 0x0A, 0x00, 0x00, 0x08,
        0x84, 0x0E, 0x01, 0x0E, 0x01, 0x10,
        0x80, 0x0A, 0x0A, 0x00, 0x00, 0x08,
        0x00, 0x04, 0x10, 0x00, 0x00, 0x05,
        0x00, 0x03, 0x0E, 0x00, 0x00, 0x0A,
        0x00, 0x23, 0x00, 0x00, 0x00, 0x01
    ]

    # R22H    r
    lut_bw = [
        0xA0, 0x1A, 0x1A, 0x00, 0x00, 0x01,
        0x00, 0x0A, 0x0A, 0x00, 0x00, 0x08,
        0x84, 0x0E, 0x01, 0x0E, 0x01, 0x10,
        0x90, 0x0A, 0x0A, 0x00, 0x00, 0x08,
        0xB0, 0x04, 0x10, 0x00, 0x00, 0x05,
        0xB0, 0x03, 0x0E, 0x00, 0x00, 0x0A,
        0xC0, 0x23, 0x00, 0x00, 0x00, 0x01
    ]

    # R23H    w
    lut_bb = [
        0x90, 0x1A, 0x1A, 0x00, 0x00, 0x01,
        0x40, 0x0A, 0x0A, 0x00, 0x00, 0x08,
        0x84, 0x0E, 0x01, 0x0E, 0x01, 0x10,
        0x80, 0x0A, 0x0A, 0x00, 0x00, 0x08,
        0x00, 0x04, 0x10, 0x00, 0x00, 0x05,
        0x00, 0x03, 0x0E, 0x00, 0x00, 0x0A,
        0x00, 0x23, 0x00, 0x00, 0x00, 0x01
    ]

    # R24H    b
    lut_wb = [
        0x90, 0x1A, 0x1A, 0x00, 0x00, 0x01,
        0x20, 0x0A, 0x0A, 0x00, 0x00, 0x08,
        0x84, 0x0E, 0x01, 0x0E, 0x01, 0x10,
        0x10, 0x0A, 0x0A, 0x00, 0x00, 0x08,
        0x00, 0x04, 0x10, 0x00, 0x00, 0x05,
        0x00, 0x03, 0x0E, 0x00, 0x00, 0x0A,
        0x00, 0x23, 0x00, 0x00, 0x00, 0x01
    ]

    def __init__(self):
        super().__init__(name='2.7" B', width=176, height=264)

    def init(self, **kwargs):
        if self.epd_init() != 0:
            return -1
        self.reset()

        self.send_command(self.POWER_ON)
        self.wait_until_idle()

        self.send_command(self.PANEL_SETTING)
        self.send_data(0xaf)  # KW-BF   KWR-AF    BWROTP 0f

        self.send_command(self.PLL_CONTROL)
        self.send_data(0x3a)  # 3A 100HZ   29 150Hz 39 200HZ    31 171HZ

        self.send_command(self.POWER_SETTING)
        self.send_data(0x03)  # VDS_EN, VDG_EN
        self.send_data(0x00)  # VCOM_HV, VGHL_LV[1], VGHL_LV[0]
        self.send_data(0x2b)  # VDH
        self.send_data(0x2b)  # VDL
        self.send_data(0x09)  # VDHR

        self.send_command(self.BOOSTER_SOFT_START)
        self.send_data(0x07)
        self.send_data(0x07)
        self.send_data(0x17)

        # Power optimization
        self.send_command(0xF8)
        self.send_data(0x60)
        self.send_data(0xA5)

        # Power optimization
        self.send_command(0xF8)
        self.send_data(0x89)
        self.send_data(0xA5)

        # Power optimization
        self.send_command(0xF8)
        self.send_data(0x90)
        self.send_data(0x00)

        # Power optimization
        self.send_command(0xF8)
        self.send_data(0x93)
        self.send_data(0x2A)

        # Power optimization
        self.send_command(0xF8)
        self.send_data(0x73)
        self.send_data(0x41)

        self.send_command(self.VCM_DC_SETTING)
        self.send_data(0x12)
        self.send_command(self.VCOM_AND_DATA_INTERVAL_SETTING)
        self.send_data(0x87)  # define by OTP

        self.set_lut()

        self.send_command(self.PARTIAL_DISPLAY_REFRESH)
        self.send_data(0x00)

        return 0

    def set_lut(self):
        self.send_command(self.LUT_FOR_VCOM)  # vcom
        for count in range(0, 44):
            self.send_data(self.lut_vcom_dc[count])

        self.send_command(self.LUT_WHITE_TO_WHITE)  # ww --
        for count in range(0, 42):
            self.send_data(self.lut_ww[count])

        self.send_command(self.LUT_BLACK_TO_WHITE)  # bw r
        for count in range(0, 42):
            self.send_data(self.lut_bw[count])

        self.send_command(self.LUT_WHITE_TO_BLACK)  # wb w
        for count in range(0, 42):
            self.send_data(self.lut_bb[count])

        self.send_command(self.LUT_BLACK_TO_BLACK)  # bb b
        for count in range(0, 42):
            self.send_data(self.lut_wb[count])

    def get_frame_buffer(self, image, reverse=True):
        super().get_frame_buffer(image, reverse=reverse)

    def display_frame(self, frame_buffer_black, *args):
        frame_buffer_red = args[0] if args else None
        self.send_command(self.TCON_RESOLUTION)
        self.send_data(self.EPD_WIDTH >> 8)
        self.send_data(self.EPD_WIDTH & 0xff)  # 176
        self.send_data(self.EPD_HEIGHT >> 8)
        self.send_data(self.EPD_HEIGHT & 0xff)  # 264

        if frame_buffer_black:
            self.send_command(self.DATA_START_TRANSMISSION_1)
            self.delay_ms(2)
            for i in range(0, int(self.width * self.height / 8)):
                self.send_data(frame_buffer_black[i])
            self.delay_ms(2)
        if frame_buffer_red:
            self.send_command(self.DATA_START_TRANSMISSION_2)
            self.delay_ms(2)
            for i in range(0, int(self.width * self.height / 8)):
                self.send_data(frame_buffer_red[i])
            self.delay_ms(2)

        self.send_command(self.DISPLAY_REFRESH)
        self.wait_until_idle()

    def set_absolute_pixel(self, frame_buffer, x, y, colored, reverse=True):
        super().set_absolute_pixel(frame_buffer, x, y, colored, reverse=reverse)

    # After this command is transmitted, the chip would enter the deep-sleep
    # mode to save power. The deep sleep mode would return to standby by
    # hardware reset. The only one parameter is a check code, the command would
    # be executed if check code = 0xA5.
    # Use EPD::Reset() to awaken and use EPD::Init() to initialize.
    def sleep(self, sleepbyte=None):
        self.send_command(self.DEEP_SLEEP)
        self.send_data(0xa5)


class EPD2in9b(WaveshareColorDraw):
    """Waveshare 2.9" B - black / white / red"""

    PARTIAL_IN = 0x91
    PARTIAL_OUT = 0x92
    PARTIAL_WINDOW = 0x90
    POWER_SAVING = 0xE3
    TCON_RESOLUTION = 0x61
    TEMPERATURE_SENSOR_CALIBRATION = 0x41

    def __init__(self):
        super().__init__(name='2.9" B', width=128, height=296)

    def init(self, **kwargs):
        if self.epd_init() != 0:
            return -1
        self.reset()
        self.send_command(self.BOOSTER_SOFT_START)
        self.send_data(0x17)
        self.send_data(0x17)
        self.send_data(0x17)
        self.send_command(self.POWER_ON)
        self.wait_until_idle()
        self.send_command(self.PANEL_SETTING)
        self.send_data(0x8F)
        self.send_command(self.VCOM_AND_DATA_INTERVAL_SETTING)
        self.send_data(0x77)
        self.send_command(self.TCON_RESOLUTION)
        self.send_data(0x80)
        self.send_data(0x01)
        self.send_data(0x28)
        self.send_command(self.VCM_DC_SETTING)
        self.send_data(0x0A)

    def get_frame_buffer(self, image, reverse=True):
        super().get_frame_buffer(image, reverse=reverse)

    def sleep(self, sleepbyte=0x37):
        super().sleep(sleepbyte=sleepbyte)



from PIL import Image
import array
import struct
import time

from papertty.drivers.drivers_base import DisplayDriver

try:
    import spidev
    import RPi.GPIO as GPIO
except ImportError:
    pass
except RuntimeError as e:
    print(str(e))


class IT8951(DisplayDriver):
    """A generic driver for displays that use a IT8951 controller board.

    This class will automatically infer the width and height by querying the
    controller."""

    RST_PIN = 17
    CS_PIN = 8
    BUSY_PIN = 24

    VCOM = 2000

    CMD_GET_DEVICE_INFO = [0x03, 0x02]
    CMD_WRITE_REGISTER = [0x00, 0x11]
    CMD_READ_REGISTER = [0x00, 0x10]
    CMD_DISPLAY_AREA = [0x00, 0x34]
    CMD_VCOM = [0x00, 0x39]
    CMD_LOAD_IMAGE_AREA = [0x00, 0x21]
    CMD_LOAD_IMAGE_END = [0x00, 0x22]

    REG_SYSTEM_BASE = 0
    REG_I80CPCR = REG_SYSTEM_BASE + 0x04

    REG_DISPLAY_BASE = 0x1000
    REG_LUTAFSR = REG_DISPLAY_BASE + 0x224 # LUT Status Reg (status of All LUT Engines)

    REG_MEMORY_CONV_BASE_ADDR = 0x0200
    REG_MEMORY_CONV = REG_MEMORY_CONV_BASE_ADDR + 0x0000
    REG_MEMORY_CONV_LISAR = REG_MEMORY_CONV_BASE_ADDR + 0x0008

    ROTATE_0   = 0
    ROTATE_90  = 1
    ROTATE_180 = 2
    ROTATE_270 = 3

    BPP_2 = 0
    BPP_3 = 1
    BPP_4 = 2
    BPP_8 = 3

    LOAD_IMAGE_L_ENDIAN = 0
    LOAD_IMAGE_B_ENDIAN = 1

    # Erases the display and leaves it in a white state regardless of what image
    # data is currently in memory.
    DISPLAY_UPDATE_MODE_INIT = 0
    # A fast non-flashy update mode that can go from any gray scale color to
    # black or white.
    DISPLAY_UPDATE_MODE_DU   = 1
    # A flashy update mode that can go from any gray scale color to any other
    # gray scale color.
    DISPLAY_UPDATE_MODE_GC16 = 2
    # For more documentation on display update modes see the reference document:
    # http://www.waveshare.net/w/upload/c/c4/E-paper-mode-declaration.pdf

    def __init__(self):
        super().__init__()
        self.name = "IT8951"
        self.supports_partial = True

    def delay_ms(self, delaytime):
        time.sleep(float(delaytime) / 1000.0)

    def spi_write(self, data):
        """Write raw bytes over SPI."""
        self.SPI.writebytes(data)

    def spi_read(self, n):
        """Read n raw bytes over SPI."""
        return self.SPI.readbytes(n)

    def write_command(self, command):
        self.wait_for_ready()
        GPIO.output(self.CS_PIN, GPIO.LOW)
        self.spi_write([0x60, 0x00])
        self.wait_for_ready()
        self.spi_write(command)
        GPIO.output(self.CS_PIN, GPIO.HIGH)

    def write_data_bytes(self, data):
        max_transfer_size = 4096
        self.wait_for_ready()
        GPIO.output(self.CS_PIN, GPIO.LOW)
        self.spi_write([0x00, 0x00])
        self.wait_for_ready()
        for i in range(0, len(data), max_transfer_size):
            self.spi_write(data[i: i + max_transfer_size])
        GPIO.output(self.CS_PIN, GPIO.HIGH)

    def read_bytes(self, n):
        self.wait_for_ready()
        GPIO.output(self.CS_PIN, GPIO.LOW)
        self.spi_write([0x10, 0x00])
        self.wait_for_ready()
        self.spi_write([0x00, 0x00]) # Two bytes of dummy data.
        self.wait_for_ready()
        result = array.array("B", self.spi_read(n))
        GPIO.output(self.CS_PIN, GPIO.HIGH)
        return result

    def write_data_half_word(self, half_word):
        """Writes a half word of data to the controller.

        The standard integer format for passing data to and from the controller
        is little endian 16 bit words.
        """
        self.write_data_bytes([(half_word >> 8) & 0xFF, half_word & 0xFF])

    def read_half_word(self):
        """Reads a half word of from the controller."""
        return struct.unpack(">H", self.read_bytes(2))[0]

    def write_register(self, register_address, value):
        self.write_command(self.CMD_WRITE_REGISTER)
        self.write_data_half_word(register_address)
        self.write_data_half_word(value)

    def read_register(self, register_address):
        self.write_command(self.CMD_READ_REGISTER)
        self.write_data_half_word(register_address)
        return self.read_half_word()

    def wait_for_ready(self):
        """Waits for the busy pin to drop.

        When the busy pin is high the controller is busy and may drop any
        commands that are sent to it."""
        while GPIO.input(self.BUSY_PIN) == 0:
            self.delay_ms(100)

    def wait_for_display_ready(self):
        """Waits for the display to be finished updating.

        It is possible for the controller to be ready for more commands but the
        display to still be refreshing. This will wait for the display to be
        stable."""
        while self.read_register(self.REG_LUTAFSR) != 0:
            self.delay_ms(100)

    def get_vcom(self):
        self.wait_for_ready()
        self.write_command(self.CMD_VCOM)
        self.write_data_half_word(0)
        return self.read_half_word()

    def set_vcom(self, vcom):
        self.write_command(self.CMD_VCOM)
        self.write_data_half_word(1)
        self.write_data_half_word(vcom)

    def fixup_string(self, s):
        result = ""
        for i in range(0, len(s), 2):
            result += "%c%c" % (s[i + 1], s[i])
        null_index = result.find("\0")
        if null_index != -1:
            result = result[0:null_index]
        return result

    def init(self, **kwargs):
        GPIO.setmode(GPIO.BCM)
        GPIO.setwarnings(False)
        GPIO.setup(self.RST_PIN, GPIO.OUT)
        GPIO.setup(self.CS_PIN, GPIO.OUT)
        GPIO.setup(self.BUSY_PIN, GPIO.IN)
        self.SPI = spidev.SpiDev(0, 0)
        self.SPI.max_speed_hz = 2000000
        self.SPI.mode = 0b00

        # It is unclear why this is necessary but it appears to be. The sample
        # code from WaveShare [1] manually controls the CS bin and has its state
        # span multiple SPI operations.
        #
        # [1] https://github.com/waveshare/IT8951
        self.SPI.no_cs = True

        GPIO.output(self.CS_PIN, GPIO.HIGH)

        # Reset the device to its initial state.
        GPIO.output(self.RST_PIN, GPIO.LOW)
        self.delay_ms(500)
        GPIO.output(self.RST_PIN, GPIO.HIGH)
        self.delay_ms(500)

        self.write_command(self.CMD_GET_DEVICE_INFO);

        (
                self.width,
                self.height,
                img_addr_l,
                img_addr_h,
                firmware_version,
                lut_version,
        ) = struct.unpack(">HHHH16s16s", self.read_bytes(40))
        firmware_version = self.fixup_string(firmware_version)
        lut_version = self.fixup_string(lut_version)
        self.img_addr = img_addr_h << 16 | img_addr_l

        print("width = %d" % self.width)
        print("height = %d" % self.height)
        print("img_addr = %08x" % self.img_addr)
        print("firmware = %s" % firmware_version)
        print("lut = %s" % lut_version)

        # Ensure that the returned device info looks sane. If it doesn't, then
        # there is little chance that any of the other operations are going to
        # do anything.
        assert self.img_addr != 0
        assert self.width != 0
        assert self.height != 0

        # Set to Enable I80 Packed mode.
        self.write_register(self.REG_I80CPCR, 0x0001)

        if self.VCOM != self.get_vcom():
            self.set_vcom(self.VCOM)
            print("VCOM = -%.02fV" % (self.get_vcom() / 1000.0))

        # Initialize the display with a blank image.
        self.wait_for_ready()
        image = Image.new("L", (self.width, self.height), 0x255)
        self.draw(0, 0, image, self.DISPLAY_UPDATE_MODE_INIT)

    def display_area(self, x, y, w, h, display_mode):
        self.write_command(self.CMD_DISPLAY_AREA)
        self.write_data_half_word(x)
        self.write_data_half_word(y)
        self.write_data_half_word(w)
        self.write_data_half_word(h)
        self.write_data_half_word(display_mode)

    def draw(self, x, y, image, update_mode_override=None):
        width = image.size[0]
        height = image.size[1]

        self.wait_for_display_ready()

        self.write_register(
                self.REG_MEMORY_CONV_LISAR + 2, (self.img_addr >> 16) & 0xFFFF)
        self.write_register(self.REG_MEMORY_CONV_LISAR, self.img_addr & 0xFFFF)

        # Define the region being loaded.
        self.write_command(self.CMD_LOAD_IMAGE_AREA)
        self.write_data_half_word(
                (self.LOAD_IMAGE_L_ENDIAN << 8) |
                (self.BPP_4 << 4) |
                self.ROTATE_0)
        self.write_data_half_word(x)
        self.write_data_half_word(y)
        self.write_data_half_word(width)
        self.write_data_half_word(height)

        packed_image = self.pack_image(image)
        self.write_data_bytes(packed_image)
        self.write_command(self.CMD_LOAD_IMAGE_END);

        if update_mode_override is not None:
            update_mode = update_mode_override
        elif image.mode == "1":
            # Use a faster, non-flashy update mode for pure black and white
            # images.
            update_mode = self.DISPLAY_UPDATE_MODE_DU
        else:
            # Use a slower, flashy update mode for gray scale images.
            update_mode = self.DISPLAY_UPDATE_MODE_GC16
        # Blit the image to the display
        self.display_area(x, y, width, height, update_mode)

    def clear(self):
        image = Image.new('1', (self.width, self.height), self.white)
        self.draw(0, 0, image, self.DISPLAY_UPDATE_MODE_INIT)

    def pack_image(self, image):
        """Packs a PIL image for transfer over SPI to the driver board."""
        if image.mode == '1':
            # B/W pictured can be processed more quickly
            frame_buffer = list(image.getdata())

            # For now, only 4 bit packing is supported. Theoretically we could
            # achieve a transfer speed up by using 2 bit packing for black and white
            # images. However, 2bpp doesn't seem to play well with the DU rendering
            # mode.
            packed_buffer = []
            # The driver board assumes all data is read in as 16bit ints. To match
            # the endianness every pair of bytes must be swapped.
            # The image is always padded to a multiple of 8, so we can safely go in steps of 4.
            for i in range(0, len(frame_buffer), 4):
                if frame_buffer[i + 2] and frame_buffer[i + 3]:
                    packed_buffer += [0xFF]
                elif frame_buffer[i + 2]:
                    packed_buffer += [0x0F]
                elif frame_buffer[i + 3]:
                    packed_buffer += [0xF0]
                else:
                    packed_buffer += [0]

                if frame_buffer[i] and frame_buffer[i + 1]:
                    packed_buffer += [0xFF]
                elif frame_buffer[i]:
                    packed_buffer += [0x0F]
                elif frame_buffer[i + 1]:
                    packed_buffer += [0xF0]
                else:
                    packed_buffer += [0]
            return packed_buffer
        else:
            # old packing code for grayscale (VNC)
            image_grey = image.convert("L")
            frame_buffer = list(image_grey.getdata())

            # For now, only 4 bit packing is supported. Theoretically we could
            # achieve a transfer speed up by using 2 bit packing for black and white
            # images. However, 2bpp doesn't seem to play well with the DU rendering
            # mode.
            packed_buffer = []

            # The driver board assumes all data is read in as 16bit ints. To match
            # the endianness every pair of bytes must be swapped.
            # The image is always padded to a multiple of 8, so we can safely go in steps of 4.
            for i in range(0, len(frame_buffer), 4):
                # Values are in the range 0..255, so we don't need to "and" after we shift
                packed_buffer += [(frame_buffer[i + 2] >> 4) | (frame_buffer[i + 3] & 0xF0)]
                packed_buffer += [(frame_buffer[i] >> 4) | (frame_buffer[i + 1] & 0xF0)]

            return packed_buffer
